# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""The spm module provides basic functions for interfacing with matlab
and spm to access spm tools.
"""
# Standard library imports
import os
from glob import glob

# Third-party imports
import numpy as np

# Local imports
from ... import logging
from ...utils.filemanip import ensure_list, simplify_list, split_filename
from ..base import (
    Bunch,
    traits,
    TraitedSpec,
    File,
    Directory,
    OutputMultiPath,
    InputMultiPath,
    isdefined,
)
from .base import SPMCommand, SPMCommandInputSpec, scans_for_fnames, ImageFileSPM

__docformat__ = "restructuredtext"
iflogger = logging.getLogger("nipype.interface")


class Level1DesignInputSpec(SPMCommandInputSpec):
    spm_mat_dir = Directory(
        exists=True, field="dir", desc="directory to store SPM.mat file (opt)"
    )
    timing_units = traits.Enum(
        "secs",
        "scans",
        field="timing.units",
        desc="units for specification of onsets",
        mandatory=True,
    )
    interscan_interval = traits.Float(
        field="timing.RT", desc="Interscan interval in secs", mandatory=True
    )
    microtime_resolution = traits.Int(
        field="timing.fmri_t", desc=("Number of time-bins per scan in secs (opt)")
    )
    microtime_onset = traits.Float(
        field="timing.fmri_t0",
        desc=("The onset/time-bin in seconds for alignment (opt)"),
    )
    session_info = traits.Any(
        field="sess",
        desc=("Session specific information generated by ``modelgen.SpecifyModel``"),
        mandatory=True,
    )
    factor_info = traits.List(
        traits.Dict(traits.Enum("name", "levels")),
        field="fact",
        desc=("Factor specific information file (opt)"),
    )
    bases = traits.Dict(
        traits.Enum("hrf", "fourier", "fourier_han", "gamma", "fir"),
        field="bases",
        desc="""\
Dictionary names of the basis function to parameters:

    * hrf

        * derivs -- (2-element list) Model  HRF  Derivatives. No derivatives: [0,0],
          Time derivatives : [1,0], Time and Dispersion derivatives: [1,1]

    * fourier, fourier_han, gamma, or fir:

        * length -- (int) Post-stimulus window length (in seconds)
        * order -- (int) Number of basis functions

""",
        mandatory=True,
    )
    volterra_expansion_order = traits.Enum(
        1, 2, field="volt", desc=("Model interactions - no:1, yes:2")
    )
    global_intensity_normalization = traits.Enum(
        "none",
        "scaling",
        field="global",
        desc=("Global intensity normalization - scaling or none"),
    )
    mask_image = File(
        exists=True, field="mask", desc="Image  for  explicitly  masking the analysis"
    )
    mask_threshold = traits.Either(
        traits.Enum("-Inf"),
        traits.Float(),
        desc="Thresholding for the mask",
        default="-Inf",
        usedefault=True,
    )
    model_serial_correlations = traits.Enum(
        "AR(1)",
        "FAST",
        "none",
        field="cvi",
        desc=(
            "Model serial correlations "
            "AR(1), FAST or none. FAST "
            "is available in SPM12"
        ),
    )
    flags = traits.Dict(
        desc="Additional arguments to the job, e.g., a common SPM operation is to "
        "modify the default masking threshold (mthresh)"
    )


class Level1DesignOutputSpec(TraitedSpec):
    spm_mat_file = File(exists=True, desc="SPM mat file")


class Level1Design(SPMCommand):
    """Generate an SPM design matrix

    http://www.fil.ion.ucl.ac.uk/spm/doc/manual.pdf#page=59

    Examples
    --------

    >>> level1design = Level1Design()
    >>> level1design.inputs.timing_units = 'secs'
    >>> level1design.inputs.interscan_interval = 2.5
    >>> level1design.inputs.bases = {'hrf':{'derivs': [0,0]}}
    >>> level1design.inputs.session_info = 'session_info.npz'
    >>> level1design.inputs.flags = {'mthresh': 0.4}
    >>> level1design.run() # doctest: +SKIP

    """

    input_spec = Level1DesignInputSpec
    output_spec = Level1DesignOutputSpec

    _jobtype = "stats"
    _jobname = "fmri_spec"

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for spm
        """
        if opt in ["spm_mat_dir", "mask_image"]:
            return np.array([str(val)], dtype=object)
        if opt in ["session_info"]:  # , 'factor_info']:
            if isinstance(val, dict):
                return [val]
            else:
                return val
        return super(Level1Design, self)._format_arg(opt, spec, val)

    def _parse_inputs(self):
        """validate spm realign options if set to None ignore
        """
        einputs = super(Level1Design, self)._parse_inputs(
            skip=("mask_threshold", "flags")
        )
        if isdefined(self.inputs.flags):
            einputs[0].update({flag: val for (flag, val) in self.inputs.flags.items()})
        for sessinfo in einputs[0]["sess"]:
            sessinfo["scans"] = scans_for_fnames(
                ensure_list(sessinfo["scans"]), keep4d=False
            )
        if not isdefined(self.inputs.spm_mat_dir):
            einputs[0]["dir"] = np.array([str(os.getcwd())], dtype=object)
        return einputs

    def _make_matlab_command(self, content):
        """validates spm options and generates job structure
        if mfile is True uses matlab .m file
        else generates a job structure and saves in .mat
        """
        if isdefined(self.inputs.mask_image):
            # SPM doesn't handle explicit masking properly, especially
            # when you want to use the entire mask image
            postscript = "load SPM;\n"
            postscript += "SPM.xM.VM = spm_vol('%s');\n" % simplify_list(
                self.inputs.mask_image
            )
            postscript += "SPM.xM.I = 0;\n"
            postscript += "SPM.xM.T = [];\n"
            postscript += (
                "SPM.xM.TH = ones(size(SPM.xM.TH))*(%s);\n" % self.inputs.mask_threshold
            )
            postscript += "SPM.xM.xs = struct('Masking', 'explicit masking only');\n"
            postscript += "save SPM SPM;\n"
        else:
            postscript = None
        return super(Level1Design, self)._make_matlab_command(
            content, postscript=postscript
        )

    def _list_outputs(self):
        outputs = self._outputs().get()
        spm = os.path.join(os.getcwd(), "SPM.mat")
        outputs["spm_mat_file"] = spm
        return outputs


class EstimateModelInputSpec(SPMCommandInputSpec):
    spm_mat_file = File(
        exists=True,
        field="spmmat",
        copyfile=True,
        mandatory=True,
        desc="Absolute path to SPM.mat",
    )
    estimation_method = traits.Dict(
        traits.Enum("Classical", "Bayesian2", "Bayesian"),
        field="method",
        mandatory=True,
        desc=("Dictionary of either Classical: 1, Bayesian: 1, or Bayesian2: 1 (dict)"),
    )
    write_residuals = traits.Bool(
        field="write_residuals", desc="Write individual residual images"
    )
    flags = traits.Dict(desc="Additional arguments")


class EstimateModelOutputSpec(TraitedSpec):
    mask_image = ImageFileSPM(exists=True, desc="binary mask to constrain estimation")
    beta_images = OutputMultiPath(
        ImageFileSPM(exists=True), desc="design parameter estimates"
    )
    residual_image = ImageFileSPM(
        exists=True, desc="Mean-squared image of the residuals"
    )
    residual_images = OutputMultiPath(
        ImageFileSPM(exists=True),
        desc="individual residual images (requires `write_residuals`",
    )
    RPVimage = ImageFileSPM(exists=True, desc="Resels per voxel image")
    spm_mat_file = File(exists=True, desc="Updated SPM mat file")
    labels = ImageFileSPM(exists=True, desc="label file")
    SDerror = OutputMultiPath(
        ImageFileSPM(exists=True), desc="Images of the standard deviation of the error"
    )
    ARcoef = OutputMultiPath(
        ImageFileSPM(exists=True), desc="Images of the AR coefficient"
    )
    Cbetas = OutputMultiPath(
        ImageFileSPM(exists=True), desc="Images of the parameter posteriors"
    )
    SDbetas = OutputMultiPath(
        ImageFileSPM(exists=True),
        desc="Images of the standard deviation of parameter posteriors",
    )


class EstimateModel(SPMCommand):
    """Use spm_spm to estimate the parameters of a model

    http://www.fil.ion.ucl.ac.uk/spm/doc/manual.pdf#page=69

    Examples
    --------
    >>> est = EstimateModel()
    >>> est.inputs.spm_mat_file = 'SPM.mat'
    >>> est.inputs.estimation_method = {'Classical': 1}
    >>> est.run() # doctest: +SKIP
    """

    input_spec = EstimateModelInputSpec
    output_spec = EstimateModelOutputSpec
    _jobtype = "stats"
    _jobname = "fmri_est"

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for spm
        """
        if opt == "spm_mat_file":
            return np.array([str(val)], dtype=object)
        if opt == "estimation_method":
            if isinstance(val, (str, bytes)):
                return {"{}".format(val): 1}
            else:
                return val
        return super(EstimateModel, self)._format_arg(opt, spec, val)

    def _parse_inputs(self):
        """validate spm realign options if set to None ignore
        """
        einputs = super(EstimateModel, self)._parse_inputs(skip=("flags"))
        if isdefined(self.inputs.flags):
            einputs[0].update({flag: val for (flag, val) in self.inputs.flags.items()})
        return einputs

    def _list_outputs(self):
        import scipy.io as sio

        outputs = self._outputs().get()
        pth = os.path.dirname(self.inputs.spm_mat_file)
        outtype = "nii" if "12" in self.version.split(".")[0] else "img"
        spm = sio.loadmat(self.inputs.spm_mat_file, struct_as_record=False)

        betas = [vbeta.fname[0] for vbeta in spm["SPM"][0, 0].Vbeta[0]]
        if (
            "Bayesian" in self.inputs.estimation_method.keys()
            or "Bayesian2" in self.inputs.estimation_method.keys()
        ):
            outputs["labels"] = os.path.join(pth, "labels.{}".format(outtype))
            outputs["SDerror"] = glob(os.path.join(pth, "Sess*_SDerror*"))
            outputs["ARcoef"] = glob(os.path.join(pth, "Sess*_AR_*"))
            if betas:
                outputs["Cbetas"] = [
                    os.path.join(pth, "C{}".format(beta)) for beta in betas
                ]
                outputs["SDbetas"] = [
                    os.path.join(pth, "SD{}".format(beta)) for beta in betas
                ]

        if "Classical" in self.inputs.estimation_method.keys():
            outputs["residual_image"] = os.path.join(pth, "ResMS.{}".format(outtype))
            outputs["RPVimage"] = os.path.join(pth, "RPV.{}".format(outtype))
            if self.inputs.write_residuals:
                outputs["residual_images"] = glob(os.path.join(pth, "Res_*"))
            if betas:
                outputs["beta_images"] = [os.path.join(pth, beta) for beta in betas]

        outputs["mask_image"] = os.path.join(pth, "mask.{}".format(outtype))
        outputs["spm_mat_file"] = os.path.join(pth, "SPM.mat")
        return outputs


class EstimateContrastInputSpec(SPMCommandInputSpec):
    spm_mat_file = File(
        exists=True,
        field="spmmat",
        desc="Absolute path to SPM.mat",
        copyfile=True,
        mandatory=True,
    )
    contrasts = traits.List(
        traits.Either(
            traits.Tuple(
                traits.Str,
                traits.Enum("T"),
                traits.List(traits.Str),
                traits.List(traits.Float),
            ),
            traits.Tuple(
                traits.Str,
                traits.Enum("T"),
                traits.List(traits.Str),
                traits.List(traits.Float),
                traits.List(traits.Float),
            ),
            traits.Tuple(
                traits.Str,
                traits.Enum("F"),
                traits.List(
                    traits.Either(
                        traits.Tuple(
                            traits.Str,
                            traits.Enum("T"),
                            traits.List(traits.Str),
                            traits.List(traits.Float),
                        ),
                        traits.Tuple(
                            traits.Str,
                            traits.Enum("T"),
                            traits.List(traits.Str),
                            traits.List(traits.Float),
                            traits.List(traits.Float),
                        ),
                    )
                ),
            ),
        ),
        desc="""List of contrasts with each contrast being a list of the form:
            [('name', 'stat', [condition list], [weight list], [session list])]
            If session list is None or not provided, all sessions are used. For
            F contrasts, the condition list should contain previously defined
            T-contrasts.""",
        mandatory=True,
    )
    beta_images = InputMultiPath(
        File(exists=True),
        desc=("Parameter estimates of the design matrix"),
        copyfile=False,
        mandatory=True,
    )
    residual_image = File(
        exists=True,
        desc="Mean-squared image of the residuals",
        copyfile=False,
        mandatory=True,
    )
    use_derivs = traits.Bool(
        desc="use derivatives for estimation", xor=["group_contrast"]
    )
    group_contrast = traits.Bool(desc="higher level contrast", xor=["use_derivs"])


class EstimateContrastOutputSpec(TraitedSpec):
    con_images = OutputMultiPath(
        File(exists=True), desc="contrast images from a t-contrast"
    )
    spmT_images = OutputMultiPath(
        File(exists=True), desc="stat images from a t-contrast"
    )
    ess_images = OutputMultiPath(
        File(exists=True), desc="contrast images from an F-contrast"
    )
    spmF_images = OutputMultiPath(
        File(exists=True), desc="stat images from an F-contrast"
    )
    spm_mat_file = File(exists=True, desc="Updated SPM mat file")


class EstimateContrast(SPMCommand):
    """Use spm_contrasts to estimate contrasts of interest

    Examples
    --------
    >>> import nipype.interfaces.spm as spm
    >>> est = spm.EstimateContrast()
    >>> est.inputs.spm_mat_file = 'SPM.mat'
    >>> cont1 = ('Task>Baseline','T', ['Task-Odd','Task-Even'],[0.5,0.5])
    >>> cont2 = ('Task-Odd>Task-Even','T', ['Task-Odd','Task-Even'],[1,-1])
    >>> contrasts = [cont1,cont2]
    >>> est.inputs.contrasts = contrasts
    >>> est.run() # doctest: +SKIP

    """

    input_spec = EstimateContrastInputSpec
    output_spec = EstimateContrastOutputSpec
    _jobtype = "stats"
    _jobname = "con"

    def _make_matlab_command(self, _):
        """Validate spm options and generate job structure."""
        contrasts = []
        cname = []
        for i, cont in enumerate(self.inputs.contrasts):
            cname.insert(i, cont[0])
            contrasts.insert(
                i,
                Bunch(
                    name=cont[0],
                    stat=cont[1],
                    conditions=cont[2],
                    weights=None,
                    sessions=None,
                ),
            )
            if len(cont) >= 4:
                contrasts[i].weights = cont[3]
            if len(cont) >= 5:
                contrasts[i].sessions = cont[4]
        script = [
            """\
%% generated by nipype.interfaces.spm
spm_defaults;
jobs{1}.stats{1}.con.spmmat  = {'%s'};
load(jobs{1}.stats{1}.con.spmmat{:});
SPM.swd = '%s';
save(jobs{1}.stats{1}.con.spmmat{:},'SPM');
names = SPM.xX.name;"""
            % (self.inputs.spm_mat_file, os.getcwd())
        ]
        # get names for columns
        if isdefined(self.inputs.group_contrast) and self.inputs.group_contrast:
            script += ["condnames=names;"]
        else:
            if self.inputs.use_derivs:
                script += [r"pat = 'Sn\([0-9]*\) (.*)';"]
            else:
                script += [
                    r"pat = 'Sn\([0-9]*\) (.*)\*bf\(1\)|Sn\([0-9]*\) "
                    r".*\*bf\([2-9]\)|Sn\([0-9]*\) (.*)';"
                ]

            script += ["t = regexp(names,pat,'tokens');"]
            # get sessidx for columns
            script += [r"pat1 = 'Sn\(([0-9].*)\)\s.*';"]
            script += ["t1 = regexp(names,pat1,'tokens');"]
            script += [
                """\
for i0=1:numel(t)
  condnames{i0}='';
  condsess(i0)=0;
  if ~isempty(t{i0}{1})
    condnames{i0} = t{i0}{1}{1};
    condsess(i0)=str2num(t1{i0}{1}{1});
  end;
end;"""
            ]

        # BUILD CONTRAST SESSION STRUCTURE
        for i, contrast in enumerate(contrasts):
            if contrast.stat == "T":
                script += ["consess{%d}.tcon.name   = '%s';" % (i + 1, contrast.name)]
                script += ["consess{%d}.tcon.convec = zeros(1,numel(names));" % (i + 1)]
                for c0, cond in enumerate(contrast.conditions):
                    script += ["idx = strmatch('%s',condnames,'exact');" % cond]
                    script += [
                        """\
if isempty(idx)
  throw(MException('CondName:Chk', sprintf('Condition %%s not found in design','%s')));
end;"""
                        % cond
                    ]
                    if contrast.sessions:
                        for sno, sw in enumerate(contrast.sessions):
                            script += ["sidx = find(condsess(idx)==%d);" % (sno + 1)]
                            script += [
                                "consess{%d}.tcon.convec(idx(sidx)) = %f;"
                                % (i + 1, sw * contrast.weights[c0],)
                            ]
                    else:
                        script += [
                            "consess{%d}.tcon.convec(idx) = %f;"
                            % (i + 1, contrast.weights[c0],)
                        ]
        for i, contrast in enumerate(contrasts):
            if contrast.stat == "F":
                script += ["consess{%d}.fcon.name   =  '%s';" % (i + 1, contrast.name)]
                for cl0, fcont in enumerate(contrast.conditions):
                    tidx = cname.index(fcont[0])
                    script += [
                        "consess{%d}.fcon.convec{%d} = consess{%d}.tcon.convec;"
                        % (i + 1, cl0 + 1, tidx + 1)
                    ]
        script += ["jobs{1}.stats{1}.con.consess = consess;"]
        script += [
            """\
if strcmp(spm('ver'),'SPM8')
  spm_jobman('initcfg');
  jobs=spm_jobman('spm5tospm8',{jobs});
end;"""
        ]
        script += ["spm_jobman('run',jobs);"]
        return "\n".join(script)

    def _list_outputs(self):
        import scipy.io as sio

        outputs = self._outputs().get()
        pth, _ = os.path.split(self.inputs.spm_mat_file)
        spm = sio.loadmat(self.inputs.spm_mat_file, struct_as_record=False)
        con_images = []
        spmT_images = []
        for con in spm["SPM"][0, 0].xCon[0]:
            con_images.append(str(os.path.join(pth, con.Vcon[0, 0].fname[0])))
            spmT_images.append(str(os.path.join(pth, con.Vspm[0, 0].fname[0])))
        if con_images:
            outputs["con_images"] = con_images
            outputs["spmT_images"] = spmT_images
        spm12 = "12" in self.version.split(".")[0]
        if spm12:
            ess = glob(os.path.join(pth, "ess*.nii"))
        else:
            ess = glob(os.path.join(pth, "ess*.img"))
        if len(ess) > 0:
            outputs["ess_images"] = sorted(ess)
        if spm12:
            spmf = glob(os.path.join(pth, "spmF*.nii"))
        else:
            spmf = glob(os.path.join(pth, "spmF*.img"))
        if len(spmf) > 0:
            outputs["spmF_images"] = sorted(spmf)
        outputs["spm_mat_file"] = self.inputs.spm_mat_file
        return outputs


class ThresholdInputSpec(SPMCommandInputSpec):
    spm_mat_file = File(
        exists=True, desc="absolute path to SPM.mat", copyfile=True, mandatory=True
    )
    stat_image = File(exists=True, desc="stat image", copyfile=False, mandatory=True)
    contrast_index = traits.Int(
        mandatory=True, desc="which contrast in the SPM.mat to use"
    )
    use_fwe_correction = traits.Bool(
        True,
        usedefault=True,
        desc=(
            "whether to use FWE (Bonferroni) "
            "correction for initial threshold "
            "(height_threshold_type has to be "
            "set to p-value)"
        ),
    )
    use_topo_fdr = traits.Bool(
        True,
        usedefault=True,
        desc=("whether to use FDR over cluster extent probabilities"),
    )
    height_threshold = traits.Float(
        0.05,
        usedefault=True,
        desc=("value for initial thresholding (defining clusters)"),
    )
    height_threshold_type = traits.Enum(
        "p-value",
        "stat",
        usedefault=True,
        desc=("Is the cluster forming threshold a stat value or p-value?"),
    )
    extent_fdr_p_threshold = traits.Float(
        0.05,
        usedefault=True,
        desc=("p threshold on FDR corrected cluster size probabilities"),
    )
    extent_threshold = traits.Int(
        0, usedefault=True, desc="Minimum cluster size in voxels"
    )
    force_activation = traits.Bool(
        False,
        usedefault=True,
        desc=(
            "In case no clusters survive the "
            "topological inference step this "
            "will pick a culster with the highes "
            "sum of t-values. Use with care."
        ),
    )


class ThresholdOutputSpec(TraitedSpec):
    thresholded_map = File(exists=True)
    n_clusters = traits.Int()
    pre_topo_fdr_map = File(exists=True)
    pre_topo_n_clusters = traits.Int()
    activation_forced = traits.Bool()
    cluster_forming_thr = traits.Float()


class Threshold(SPMCommand):
    """Topological FDR thresholding based on cluster extent/size. Smoothness is
    estimated from GLM residuals but is assumed to be the same for all of the
    voxels.

    Examples
    --------

    >>> thresh = Threshold()
    >>> thresh.inputs.spm_mat_file = 'SPM.mat'
    >>> thresh.inputs.stat_image = 'spmT_0001.img'
    >>> thresh.inputs.contrast_index = 1
    >>> thresh.inputs.extent_fdr_p_threshold = 0.05
    >>> thresh.run() # doctest: +SKIP
    """

    input_spec = ThresholdInputSpec
    output_spec = ThresholdOutputSpec

    def _gen_thresholded_map_filename(self):
        _, fname, ext = split_filename(self.inputs.stat_image)
        return os.path.abspath(fname + "_thr" + ext)

    def _gen_pre_topo_map_filename(self):
        _, fname, ext = split_filename(self.inputs.stat_image)
        return os.path.abspath(fname + "_pre_topo_thr" + ext)

    def _make_matlab_command(self, _):
        script = "con_index = %d;\n" % self.inputs.contrast_index
        script += "cluster_forming_thr = %f;\n" % self.inputs.height_threshold
        if self.inputs.use_fwe_correction:
            script += "thresDesc  = 'FWE';\n"
        else:
            script += "thresDesc  = 'none';\n"

        if self.inputs.use_topo_fdr:
            script += "use_topo_fdr  = 1;\n"
        else:
            script += "use_topo_fdr  = 0;\n"

        if self.inputs.force_activation:
            script += "force_activation  = 1;\n"
        else:
            script += "force_activation  = 0;\n"
        script += (
            "cluster_extent_p_fdr_thr = %f;\n" % self.inputs.extent_fdr_p_threshold
        )
        script += "stat_filename = '%s';\n" % self.inputs.stat_image
        script += "height_threshold_type = '%s';\n" % self.inputs.height_threshold_type
        script += "extent_threshold = %d;\n" % self.inputs.extent_threshold

        script += "load %s;\n" % self.inputs.spm_mat_file
        script += """
FWHM  = SPM.xVol.FWHM;
df = [SPM.xCon(con_index).eidf SPM.xX.erdf];
STAT = SPM.xCon(con_index).STAT;
R = SPM.xVol.R;
S = SPM.xVol.S;
n = 1;

switch thresDesc
    case 'FWE'
        cluster_forming_thr = spm_uc(cluster_forming_thr,df,STAT,R,n,S);

    case 'none'
        if strcmp(height_threshold_type, 'p-value')
            cluster_forming_thr = spm_u(cluster_forming_thr^(1/n),df,STAT);
        end
end

stat_map_vol = spm_vol(stat_filename);
[stat_map_data, stat_map_XYZmm] = spm_read_vols(stat_map_vol);

Z = stat_map_data(:)';
[x,y,z] = ind2sub(size(stat_map_data),(1:numel(stat_map_data))');
XYZ = cat(1, x', y', z');

XYZth = XYZ(:, Z >= cluster_forming_thr);
Zth = Z(Z >= cluster_forming_thr);

"""
        script += (
            "spm_write_filtered(Zth,XYZth,stat_map_vol.dim',"
            "stat_map_vol.mat,'thresholded map', '%s');\n"
        ) % self._gen_pre_topo_map_filename()
        script += """
max_size = 0;
max_size_index = 0;
th_nclusters = 0;
nclusters = 0;
if isempty(XYZth)
    thresholded_XYZ = [];
    thresholded_Z = [];
else
    if use_topo_fdr
        V2R        = 1/prod(FWHM(stat_map_vol.dim > 1));
        [uc,Pc,ue] = spm_uc_clusterFDR(cluster_extent_p_fdr_thr,df,STAT,R,n,Z,XYZ,V2R,cluster_forming_thr);
    end

    voxel_labels = spm_clusters(XYZth);
    nclusters = max(voxel_labels);

    thresholded_XYZ = [];
    thresholded_Z = [];

    for i = 1:nclusters
        cluster_size = sum(voxel_labels==i);
         if cluster_size > extent_threshold && (~use_topo_fdr || (cluster_size - uc) > -1)
            thresholded_XYZ = cat(2, thresholded_XYZ, XYZth(:,voxel_labels == i));
            thresholded_Z = cat(2, thresholded_Z, Zth(voxel_labels == i));
            th_nclusters = th_nclusters + 1;
         end
        if force_activation
            cluster_sum = sum(Zth(voxel_labels == i));
            if cluster_sum > max_size
                max_size = cluster_sum;
                max_size_index = i;
            end
        end
    end
end

activation_forced = 0;
if isempty(thresholded_XYZ)
    if force_activation && max_size ~= 0
        thresholded_XYZ = XYZth(:,voxel_labels == max_size_index);
        thresholded_Z = Zth(voxel_labels == max_size_index);
        th_nclusters = 1;
        activation_forced = 1;
    else
        thresholded_Z = [0];
        thresholded_XYZ = [1 1 1]';
        th_nclusters = 0;
    end
end

fprintf('activation_forced = %d\\n',activation_forced);
fprintf('pre_topo_n_clusters = %d\\n',nclusters);
fprintf('n_clusters = %d\\n',th_nclusters);
fprintf('cluster_forming_thr = %f\\n',cluster_forming_thr);

"""
        script += (
            "spm_write_filtered(thresholded_Z,thresholded_XYZ,"
            "stat_map_vol.dim',stat_map_vol.mat,'thresholded map',"
            " '%s');\n"
        ) % self._gen_thresholded_map_filename()

        return script

    def aggregate_outputs(self, runtime=None):
        outputs = self._outputs()
        setattr(outputs, "thresholded_map", self._gen_thresholded_map_filename())
        setattr(outputs, "pre_topo_fdr_map", self._gen_pre_topo_map_filename())
        for line in runtime.stdout.split("\n"):
            if line.startswith("activation_forced = "):
                setattr(
                    outputs,
                    "activation_forced",
                    line[len("activation_forced = ") :].strip() == "1",
                )
            elif line.startswith("n_clusters = "):
                setattr(
                    outputs, "n_clusters", int(line[len("n_clusters = ") :].strip())
                )
            elif line.startswith("pre_topo_n_clusters = "):
                setattr(
                    outputs,
                    "pre_topo_n_clusters",
                    int(line[len("pre_topo_n_clusters = ") :].strip()),
                )
            elif line.startswith("cluster_forming_thr = "):
                setattr(
                    outputs,
                    "cluster_forming_thr",
                    float(line[len("cluster_forming_thr = ") :].strip()),
                )
        return outputs

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["thresholded_map"] = self._gen_thresholded_map_filename()
        outputs["pre_topo_fdr_map"] = self._gen_pre_topo_map_filename()
        return outputs


class ThresholdStatisticsInputSpec(SPMCommandInputSpec):
    spm_mat_file = File(
        exists=True, desc="absolute path to SPM.mat", copyfile=True, mandatory=True
    )
    stat_image = File(exists=True, desc="stat image", copyfile=False, mandatory=True)
    contrast_index = traits.Int(
        mandatory=True, desc="which contrast in the SPM.mat to use"
    )
    height_threshold = traits.Float(
        desc=("stat value for initial thresholding (defining clusters)"),
        mandatory=True,
    )
    extent_threshold = traits.Int(
        0, usedefault=True, desc="Minimum cluster size in voxels"
    )


class ThresholdStatisticsOutputSpec(TraitedSpec):
    voxelwise_P_Bonf = traits.Float()
    voxelwise_P_RF = traits.Float()
    voxelwise_P_uncor = traits.Float()
    voxelwise_P_FDR = traits.Float()
    clusterwise_P_RF = traits.Float()
    clusterwise_P_FDR = traits.Float()


class ThresholdStatistics(SPMCommand):
    """Given height and cluster size threshold calculate theoretical
    probabilities concerning false positives

    Examples
    --------

    >>> thresh = ThresholdStatistics()
    >>> thresh.inputs.spm_mat_file = 'SPM.mat'
    >>> thresh.inputs.stat_image = 'spmT_0001.img'
    >>> thresh.inputs.contrast_index = 1
    >>> thresh.inputs.height_threshold = 4.56
    >>> thresh.run() # doctest: +SKIP
    """

    input_spec = ThresholdStatisticsInputSpec
    output_spec = ThresholdStatisticsOutputSpec

    def _make_matlab_command(self, _):
        script = "con_index = %d;\n" % self.inputs.contrast_index
        script += "cluster_forming_thr = %f;\n" % self.inputs.height_threshold
        script += "stat_filename = '%s';\n" % self.inputs.stat_image
        script += "extent_threshold = %d;\n" % self.inputs.extent_threshold
        script += "load '%s'\n" % self.inputs.spm_mat_file
        script += """
FWHM  = SPM.xVol.FWHM;
df = [SPM.xCon(con_index).eidf SPM.xX.erdf];
STAT = SPM.xCon(con_index).STAT;
R = SPM.xVol.R;
S = SPM.xVol.S;
n = 1;

voxelwise_P_Bonf = spm_P_Bonf(cluster_forming_thr,df,STAT,S,n)
voxelwise_P_RF = spm_P_RF(1,0,cluster_forming_thr,df,STAT,R,n)

stat_map_vol = spm_vol(stat_filename);
[stat_map_data, stat_map_XYZmm] = spm_read_vols(stat_map_vol);

Z = stat_map_data(:);
Zum = Z;

        switch STAT
            case 'Z'
                VPs = (1-spm_Ncdf(Zum)).^n;
                voxelwise_P_uncor = (1-spm_Ncdf(cluster_forming_thr)).^n
            case 'T'
                VPs = (1 - spm_Tcdf(Zum,df(2))).^n;
                voxelwise_P_uncor = (1 - spm_Tcdf(cluster_forming_thr,df(2))).^n
            case 'X'
                VPs = (1-spm_Xcdf(Zum,df(2))).^n;
                voxelwise_P_uncor = (1-spm_Xcdf(cluster_forming_thr,df(2))).^n
            case 'F'
                VPs = (1 - spm_Fcdf(Zum,df)).^n;
                voxelwise_P_uncor = (1 - spm_Fcdf(cluster_forming_thr,df)).^n
        end
        VPs = sort(VPs);

voxelwise_P_FDR = spm_P_FDR(cluster_forming_thr,df,STAT,n,VPs)

V2R        = 1/prod(FWHM(stat_map_vol.dim > 1));

clusterwise_P_RF = spm_P_RF(1,extent_threshold*V2R,cluster_forming_thr,df,STAT,R,n)

[x,y,z] = ind2sub(size(stat_map_data),(1:numel(stat_map_data))');
XYZ = cat(1, x', y', z');

[u, CPs, ue] = spm_uc_clusterFDR(0.05,df,STAT,R,n,Z,XYZ,V2R,cluster_forming_thr);

clusterwise_P_FDR = spm_P_clusterFDR(extent_threshold*V2R,df,STAT,R,n,cluster_forming_thr,CPs')
"""
        return script

    def aggregate_outputs(self, runtime=None, needed_outputs=None):
        outputs = self._outputs()
        cur_output = ""
        for line in runtime.stdout.split("\n"):
            if cur_output != "" and len(line.split()) != 0:
                setattr(outputs, cur_output, float(line))
                cur_output = ""
                continue
            if len(line.split()) != 0 and line.split()[0] in [
                "clusterwise_P_FDR",
                "clusterwise_P_RF",
                "voxelwise_P_Bonf",
                "voxelwise_P_FDR",
                "voxelwise_P_RF",
                "voxelwise_P_uncor",
            ]:
                cur_output = line.split()[0]
                continue

        return outputs


class FactorialDesignInputSpec(SPMCommandInputSpec):
    spm_mat_dir = Directory(
        exists=True, field="dir", desc="directory to store SPM.mat file (opt)"
    )
    # Need to make an alias of InputMultiPath; the inputs below are not Path
    covariates = InputMultiPath(
        traits.Dict(
            key_trait=traits.Enum("vector", "name", "interaction", "centering")
        ),
        field="cov",
        desc=("covariate dictionary {vector, name, interaction, centering}"),
    )
    threshold_mask_none = traits.Bool(
        field="masking.tm.tm_none",
        xor=["threshold_mask_absolute", "threshold_mask_relative"],
        desc="do not use threshold masking",
    )
    threshold_mask_absolute = traits.Float(
        field="masking.tm.tma.athresh",
        xor=["threshold_mask_none", "threshold_mask_relative"],
        desc="use an absolute threshold",
    )
    threshold_mask_relative = traits.Float(
        field="masking.tm.tmr.rthresh",
        xor=["threshold_mask_absolute", "threshold_mask_none"],
        desc=("threshold using a proportion of the global value"),
    )
    use_implicit_threshold = traits.Bool(
        field="masking.im", desc=("use implicit mask NaNs or zeros to threshold")
    )
    explicit_mask_file = File(
        field="masking.em",  # requires cell
        desc="use an implicit mask file to threshold",
    )
    global_calc_omit = traits.Bool(
        field="globalc.g_omit",
        xor=["global_calc_mean", "global_calc_values"],
        desc="omit global calculation",
    )
    global_calc_mean = traits.Bool(
        field="globalc.g_mean",
        xor=["global_calc_omit", "global_calc_values"],
        desc="use mean for global calculation",
    )
    global_calc_values = traits.List(
        traits.Float,
        field="globalc.g_user.global_uval",
        xor=["global_calc_mean", "global_calc_omit"],
        desc="omit global calculation",
    )
    no_grand_mean_scaling = traits.Bool(
        field="globalm.gmsca.gmsca_no", desc=("do not perform grand mean scaling")
    )
    global_normalization = traits.Enum(
        1,
        2,
        3,
        field="globalm.glonorm",
        desc=("global normalization None-1, Proportional-2, ANCOVA-3"),
    )


class FactorialDesignOutputSpec(TraitedSpec):
    spm_mat_file = File(exists=True, desc="SPM mat file")


class FactorialDesign(SPMCommand):
    """Base class for factorial designs

    http://www.fil.ion.ucl.ac.uk/spm/doc/manual.pdf#page=77

    """

    input_spec = FactorialDesignInputSpec
    output_spec = FactorialDesignOutputSpec
    _jobtype = "stats"
    _jobname = "factorial_design"

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for spm
        """
        if opt in ["spm_mat_dir", "explicit_mask_file"]:
            return np.array([str(val)], dtype=object)
        if opt in ["covariates"]:
            outlist = []
            mapping = {
                "name": "cname",
                "vector": "c",
                "interaction": "iCFI",
                "centering": "iCC",
            }
            for dictitem in val:
                outdict = {}
                for key, keyval in list(dictitem.items()):
                    outdict[mapping[key]] = keyval
                outlist.append(outdict)
            return outlist
        return super(FactorialDesign, self)._format_arg(opt, spec, val)

    def _parse_inputs(self):
        """validate spm realign options if set to None ignore
        """
        einputs = super(FactorialDesign, self)._parse_inputs()
        if not isdefined(self.inputs.spm_mat_dir):
            einputs[0]["dir"] = np.array([str(os.getcwd())], dtype=object)
        return einputs

    def _list_outputs(self):
        outputs = self._outputs().get()
        spm = os.path.join(os.getcwd(), "SPM.mat")
        outputs["spm_mat_file"] = spm
        return outputs


class OneSampleTTestDesignInputSpec(FactorialDesignInputSpec):
    in_files = traits.List(
        File(exists=True),
        field="des.t1.scans",
        mandatory=True,
        minlen=2,
        desc="input files",
    )


class OneSampleTTestDesign(FactorialDesign):
    """Create SPM design for one sample t-test

    Examples
    --------

    >>> ttest = OneSampleTTestDesign()
    >>> ttest.inputs.in_files = ['cont1.nii', 'cont2.nii']
    >>> ttest.run() # doctest: +SKIP
    """

    input_spec = OneSampleTTestDesignInputSpec

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for spm
        """
        if opt in ["in_files"]:
            return np.array(val, dtype=object)
        return super(OneSampleTTestDesign, self)._format_arg(opt, spec, val)


class TwoSampleTTestDesignInputSpec(FactorialDesignInputSpec):
    # very unlikely that you will have a single image in one group, so setting
    # parameters to require at least two files in each group [SG]
    group1_files = traits.List(
        File(exists=True),
        field="des.t2.scans1",
        mandatory=True,
        minlen=2,
        desc="Group 1 input files",
    )
    group2_files = traits.List(
        File(exists=True),
        field="des.t2.scans2",
        mandatory=True,
        minlen=2,
        desc="Group 2 input files",
    )
    dependent = traits.Bool(
        field="des.t2.dept", desc=("Are the measurements dependent between levels")
    )
    unequal_variance = traits.Bool(
        field="des.t2.variance",
        desc=("Are the variances equal or unequal between groups"),
    )


class TwoSampleTTestDesign(FactorialDesign):
    """Create SPM design for two sample t-test

    Examples
    --------

    >>> ttest = TwoSampleTTestDesign()
    >>> ttest.inputs.group1_files = ['cont1.nii', 'cont2.nii']
    >>> ttest.inputs.group2_files = ['cont1a.nii', 'cont2a.nii']
    >>> ttest.run() # doctest: +SKIP
    """

    input_spec = TwoSampleTTestDesignInputSpec

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for spm
        """
        if opt in ["group1_files", "group2_files"]:
            return np.array(val, dtype=object)
        return super(TwoSampleTTestDesign, self)._format_arg(opt, spec, val)


class PairedTTestDesignInputSpec(FactorialDesignInputSpec):
    paired_files = traits.List(
        traits.List(File(exists=True), minlen=2, maxlen=2),
        field="des.pt.pair",
        mandatory=True,
        minlen=2,
        desc="List of paired files",
    )
    grand_mean_scaling = traits.Bool(
        field="des.pt.gmsca", desc="Perform grand mean scaling"
    )
    ancova = traits.Bool(
        field="des.pt.ancova", desc="Specify ancova-by-factor regressors"
    )


class PairedTTestDesign(FactorialDesign):
    """Create SPM design for paired t-test

    Examples
    --------

    >>> pttest = PairedTTestDesign()
    >>> pttest.inputs.paired_files = [['cont1.nii','cont1a.nii'],['cont2.nii','cont2a.nii']]
    >>> pttest.run() # doctest: +SKIP
    """

    input_spec = PairedTTestDesignInputSpec

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for spm
        """
        if opt in ["paired_files"]:
            return [dict(scans=np.array(files, dtype=object)) for files in val]
        return super(PairedTTestDesign, self)._format_arg(opt, spec, val)


class MultipleRegressionDesignInputSpec(FactorialDesignInputSpec):
    in_files = traits.List(
        File(exists=True),
        field="des.mreg.scans",
        mandatory=True,
        minlen=2,
        desc="List of files",
    )
    include_intercept = traits.Bool(
        True,
        field="des.mreg.incint",
        usedefault=True,
        desc="Include intercept in design",
    )
    user_covariates = InputMultiPath(
        traits.Dict(key_trait=traits.Enum("vector", "name", "centering")),
        field="des.mreg.mcov",
        desc=("covariate dictionary {vector, name, centering}"),
    )


class MultipleRegressionDesign(FactorialDesign):
    """Create SPM design for multiple regression

    Examples
    --------

    >>> mreg = MultipleRegressionDesign()
    >>> mreg.inputs.in_files = ['cont1.nii','cont2.nii']
    >>> mreg.run() # doctest: +SKIP
    """

    input_spec = MultipleRegressionDesignInputSpec

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for spm
        """
        if opt in ["in_files"]:
            return np.array(val, dtype=object)
        if opt in ["user_covariates"]:
            outlist = []
            mapping = {"name": "cname", "vector": "c", "centering": "iCC"}
            for dictitem in val:
                outdict = {}
                for key, keyval in list(dictitem.items()):
                    outdict[mapping[key]] = keyval
                outlist.append(outdict)
            return outlist
        return super(MultipleRegressionDesign, self)._format_arg(opt, spec, val)



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..legacy import GenWarpFields


def test_GenWarpFields_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        bias_field_correction=dict(
            argstr="-n 1",
        ),
        dimension=dict(
            argstr="-d %d",
            position=1,
            usedefault=True,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        force_proceed=dict(
            argstr="-f 1",
        ),
        input_image=dict(
            argstr="-i %s",
            copyfile=False,
            extensions=None,
            mandatory=True,
        ),
        inverse_warp_template_labels=dict(
            argstr="-l",
        ),
        max_iterations=dict(
            argstr="-m %s",
            sep="x",
        ),
        num_threads=dict(
            nohash=True,
            usedefault=True,
        ),
        out_prefix=dict(
            argstr="-o %s",
            usedefault=True,
        ),
        quality_check=dict(
            argstr="-q 1",
        ),
        reference_image=dict(
            argstr="-r %s",
            copyfile=True,
            extensions=None,
            mandatory=True,
        ),
        similarity_metric=dict(
            argstr="-s %s",
        ),
        transformation_model=dict(
            argstr="-t %s",
            usedefault=True,
        ),
    )
    inputs = GenWarpFields.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_GenWarpFields_outputs():
    output_map = dict(
        affine_transformation=dict(
            extensions=None,
        ),
        input_file=dict(
            extensions=None,
        ),
        inverse_warp_field=dict(
            extensions=None,
        ),
        output_file=dict(
            extensions=None,
        ),
        warp_field=dict(
            extensions=None,
        ),
    )
    outputs = GenWarpFields.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..n4itkbiasfieldcorrection import N4ITKBiasFieldCorrection


def test_N4ITKBiasFieldCorrection_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        bsplineorder=dict(
            argstr="--bsplineorder %d",
        ),
        convergencethreshold=dict(
            argstr="--convergencethreshold %f",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        histogramsharpening=dict(
            argstr="--histogramsharpening %s",
            sep=",",
        ),
        inputimage=dict(
            argstr="--inputimage %s",
            extensions=None,
        ),
        iterations=dict(
            argstr="--iterations %s",
            sep=",",
        ),
        maskimage=dict(
            argstr="--maskimage %s",
            extensions=None,
        ),
        meshresolution=dict(
            argstr="--meshresolution %s",
            sep=",",
        ),
        outputbiasfield=dict(
            argstr="--outputbiasfield %s",
            hash_files=False,
        ),
        outputimage=dict(
            argstr="--outputimage %s",
            hash_files=False,
        ),
        shrinkfactor=dict(
            argstr="--shrinkfactor %d",
        ),
        splinedistance=dict(
            argstr="--splinedistance %f",
        ),
        weightimage=dict(
            argstr="--weightimage %s",
            extensions=None,
        ),
    )
    inputs = N4ITKBiasFieldCorrection.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_N4ITKBiasFieldCorrection_outputs():
    output_map = dict(
        outputbiasfield=dict(
            extensions=None,
        ),
        outputimage=dict(
            extensions=None,
        ),
    )
    outputs = N4ITKBiasFieldCorrection.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



"""Parallel workflow execution via PBS/Torque
"""
import os
import sys

from ...interfaces.base import CommandLine
from .sgegraph import SGEGraphPlugin
from .base import logger


class PBSGraphPlugin(SGEGraphPlugin):
    """Execute using PBS/Torque

    The plugin_args input to run can be used to control the SGE execution.
    Currently supported options are:

    - template : template to use for batch job submission
    - qsub_args : arguments to be prepended to the job execution script in the
                  qsub call

    """

    _template = """
#PBS -V
"""

    def _submit_graph(self, pyfiles, dependencies, nodes):
        batch_dir, _ = os.path.split(pyfiles[0])
        submitjobsfile = os.path.join(batch_dir, "submit_jobs.sh")
        with open(submitjobsfile, "wt") as fp:
            fp.writelines("#!/usr/bin/env sh\n")
            for idx, pyscript in enumerate(pyfiles):
                node = nodes[idx]
                template, qsub_args = self._get_args(node, ["template", "qsub_args"])

                batch_dir, name = os.path.split(pyscript)
                name = ".".join(name.split(".")[:-1])
                batchscript = "\n".join(
                    (template, "%s %s" % (sys.executable, pyscript))
                )
                batchscriptfile = os.path.join(batch_dir, "batchscript_%s.sh" % name)
                with open(batchscriptfile, "wt") as batchfp:
                    batchfp.writelines(batchscript)
                    batchfp.close()
                deps = ""
                if idx in dependencies:
                    values = ["$job%05d" % jobid for jobid in dependencies[idx]]
                    if len(values):
                        deps = "-W depend=afterok:%s" % ":".join(values)
                fp.writelines(
                    "job%05d=`qsub %s %s %s`\n"
                    % (idx, deps, qsub_args, batchscriptfile)
                )
        cmd = CommandLine(
            "sh",
            environ=dict(os.environ),
            resource_monitor=False,
            terminal_output="allatonce",
        )
        cmd.inputs.args = "%s" % submitjobsfile
        cmd.run()
        logger.info("submitted all jobs to queue")



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..preprocess import FNIRT


def test_FNIRT_inputs():
    input_map = dict(
        affine_file=dict(
            argstr="--aff=%s",
            extensions=None,
        ),
        apply_inmask=dict(
            argstr="--applyinmask=%s",
            sep=",",
            xor=["skip_inmask"],
        ),
        apply_intensity_mapping=dict(
            argstr="--estint=%s",
            sep=",",
            xor=["skip_intensity_mapping"],
        ),
        apply_refmask=dict(
            argstr="--applyrefmask=%s",
            sep=",",
            xor=["skip_refmask"],
        ),
        args=dict(
            argstr="%s",
        ),
        bias_regularization_lambda=dict(
            argstr="--biaslambda=%f",
        ),
        biasfield_resolution=dict(
            argstr="--biasres=%d,%d,%d",
        ),
        config_file=dict(
            argstr="--config=%s",
        ),
        derive_from_ref=dict(
            argstr="--refderiv",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        field_file=dict(
            argstr="--fout=%s",
            hash_files=False,
        ),
        fieldcoeff_file=dict(
            argstr="--cout=%s",
        ),
        hessian_precision=dict(
            argstr="--numprec=%s",
        ),
        in_file=dict(
            argstr="--in=%s",
            extensions=None,
            mandatory=True,
        ),
        in_fwhm=dict(
            argstr="--infwhm=%s",
            sep=",",
        ),
        in_intensitymap_file=dict(
            argstr="--intin=%s",
            copyfile=False,
        ),
        inmask_file=dict(
            argstr="--inmask=%s",
            extensions=None,
        ),
        inmask_val=dict(
            argstr="--impinval=%f",
        ),
        intensity_mapping_model=dict(
            argstr="--intmod=%s",
        ),
        intensity_mapping_order=dict(
            argstr="--intorder=%d",
        ),
        inwarp_file=dict(
            argstr="--inwarp=%s",
            extensions=None,
        ),
        jacobian_file=dict(
            argstr="--jout=%s",
            hash_files=False,
        ),
        jacobian_range=dict(
            argstr="--jacrange=%f,%f",
        ),
        log_file=dict(
            argstr="--logout=%s",
            extensions=None,
            genfile=True,
            hash_files=False,
        ),
        max_nonlin_iter=dict(
            argstr="--miter=%s",
            sep=",",
        ),
        modulatedref_file=dict(
            argstr="--refout=%s",
            hash_files=False,
        ),
        out_intensitymap_file=dict(
            argstr="--intout=%s",
            hash_files=False,
        ),
        output_type=dict(),
        ref_file=dict(
            argstr="--ref=%s",
            extensions=None,
            mandatory=True,
        ),
        ref_fwhm=dict(
            argstr="--reffwhm=%s",
            sep=",",
        ),
        refmask_file=dict(
            argstr="--refmask=%s",
            extensions=None,
        ),
        refmask_val=dict(
            argstr="--imprefval=%f",
        ),
        regularization_lambda=dict(
            argstr="--lambda=%s",
            sep=",",
        ),
        regularization_model=dict(
            argstr="--regmod=%s",
        ),
        skip_implicit_in_masking=dict(
            argstr="--impinm=0",
        ),
        skip_implicit_ref_masking=dict(
            argstr="--imprefm=0",
        ),
        skip_inmask=dict(
            argstr="--applyinmask=0",
            xor=["apply_inmask"],
        ),
        skip_intensity_mapping=dict(
            argstr="--estint=0",
            xor=["apply_intensity_mapping"],
        ),
        skip_lambda_ssq=dict(
            argstr="--ssqlambda=0",
        ),
        skip_refmask=dict(
            argstr="--applyrefmask=0",
            xor=["apply_refmask"],
        ),
        spline_order=dict(
            argstr="--splineorder=%d",
        ),
        subsampling_scheme=dict(
            argstr="--subsamp=%s",
            sep=",",
        ),
        warp_resolution=dict(
            argstr="--warpres=%d,%d,%d",
        ),
        warped_file=dict(
            argstr="--iout=%s",
            extensions=None,
            genfile=True,
            hash_files=False,
        ),
    )
    inputs = FNIRT.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_FNIRT_outputs():
    output_map = dict(
        field_file=dict(
            extensions=None,
        ),
        fieldcoeff_file=dict(
            extensions=None,
        ),
        jacobian_file=dict(
            extensions=None,
        ),
        log_file=dict(
            extensions=None,
        ),
        modulatedref_file=dict(
            extensions=None,
        ),
        out_intensitymap_file=dict(),
        warped_file=dict(
            extensions=None,
        ),
    )
    outputs = FNIRT.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-

import numpy as np
import nibabel as nb

from ... import logging
from ..base import TraitedSpec, File, isdefined
from .base import DipyDiffusionInterface, DipyBaseInterfaceInputSpec

IFLOGGER = logging.getLogger("nipype.interface")


class APMQballInputSpec(DipyBaseInterfaceInputSpec):
    mask_file = File(exists=True, desc="An optional brain mask")


class APMQballOutputSpec(TraitedSpec):
    out_file = File(exists=True)


class APMQball(DipyDiffusionInterface):
    """
    Calculates the anisotropic power map

    Example
    -------

    >>> import nipype.interfaces.dipy as dipy
    >>> apm = dipy.APMQball()
    >>> apm.inputs.in_file = 'diffusion.nii'
    >>> apm.inputs.in_bvec = 'bvecs'
    >>> apm.inputs.in_bval = 'bvals'
    >>> apm.run()                                   # doctest: +SKIP
    """

    input_spec = APMQballInputSpec
    output_spec = APMQballOutputSpec

    def _run_interface(self, runtime):
        from dipy.reconst import shm
        from dipy.data import get_sphere
        from dipy.reconst.peaks import peaks_from_model

        gtab = self._get_gradient_table()

        img = nb.load(self.inputs.in_file)
        data = np.asanyarray(img.dataobj)
        affine = img.affine
        mask = None
        if isdefined(self.inputs.mask_file):
            mask = np.asanyarray(nb.load(self.inputs.mask_file).dataobj)

        # Fit it
        model = shm.QballModel(gtab, 8)
        sphere = get_sphere("symmetric724")
        peaks = peaks_from_model(
            model=model,
            data=data,
            relative_peak_threshold=0.5,
            min_separation_angle=25,
            sphere=sphere,
            mask=mask,
        )
        apm = shm.anisotropic_power(peaks.shm_coeff)
        out_file = self._gen_filename("apm")
        nb.Nifti1Image(apm.astype("float32"), affine).to_filename(out_file)
        IFLOGGER.info("APM qball image saved as %s", out_file)

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["out_file"] = self._gen_filename("apm")

        return outputs



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Information on specific functions, classes, and methods.

:Release: |version|
:Date: |today|

Top-level module API
--------------------

"""
import os
from distutils.version import LooseVersion

from .info import URL as __url__, STATUS as __status__, __version__
from .utils.config import NipypeConfig
from .utils.logger import Logging
from .refs import due
from .pkg_info import get_pkg_info as _get_pkg_info

try:
    import faulthandler

    faulthandler.enable()
except (ImportError, IOError) as e:
    pass

config = NipypeConfig()
logging = Logging(config)


class NipypeTester(object):
    def __call__(self, doctests=True, parallel=False):
        try:
            import pytest
        except ImportError:
            raise RuntimeError("py.test not installed, run: pip install pytest")
        args = []
        if not doctests:
            args.extend(["-p", "no:doctest"])
        if parallel:
            try:
                import xdist
            except ImportError:
                raise RuntimeError("pytest-xdist required for parallel run")
            args.append("-n auto")
        args.append(os.path.dirname(__file__))
        pytest.main(args=args)


test = NipypeTester()


def get_info():
    """Returns package information"""
    return _get_pkg_info(os.path.dirname(__file__))


from .pipeline import Node, MapNode, JoinNode, Workflow
from .interfaces import (
    DataGrabber,
    DataSink,
    SelectFiles,
    IdentityInterface,
    Rename,
    Function,
    Select,
    Merge,
)


def check_latest_version(raise_exception=False):
    """
    Check for the latest version of the library.

    Parameters
    ----------
    raise_exception: bool
        Raise a RuntimeError if a bad version is being used
    """
    import etelemetry

    logger = logging.getLogger("nipype.utils")
    return etelemetry.check_available_version(
        "nipy/nipype", __version__, logger, raise_exception
    )


# Run telemetry on import for interactive sessions, such as IPython, Jupyter notebooks, Python REPL
if config.getboolean("execution", "check_version"):
    import __main__

    if not hasattr(__main__, "__file__") and "NIPYPE_NO_ET" not in os.environ:
        from .interfaces.base import BaseInterface

        if BaseInterface._etelemetry_version_data is None:
            BaseInterface._etelemetry_version_data = check_latest_version()



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..utils import ParcellationStats


def test_ParcellationStats_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        aseg=dict(
            extensions=None,
            mandatory=True,
        ),
        brainmask=dict(
            extensions=None,
            mandatory=True,
        ),
        copy_inputs=dict(),
        cortex_label=dict(
            extensions=None,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        hemisphere=dict(
            argstr="%s",
            mandatory=True,
            position=-2,
        ),
        in_annotation=dict(
            argstr="-a %s",
            extensions=None,
            xor=["in_label"],
        ),
        in_cortex=dict(
            argstr="-cortex %s",
            extensions=None,
        ),
        in_label=dict(
            argstr="-l %s",
            extensions=None,
            xor=["in_annotatoin", "out_color"],
        ),
        lh_pial=dict(
            extensions=None,
            mandatory=True,
        ),
        lh_white=dict(
            extensions=None,
            mandatory=True,
        ),
        mgz=dict(
            argstr="-mgz",
        ),
        out_color=dict(
            argstr="-c %s",
            extensions=None,
            genfile=True,
            xor=["in_label"],
        ),
        out_table=dict(
            argstr="-f %s",
            extensions=None,
            genfile=True,
            requires=["tabular_output"],
        ),
        rh_pial=dict(
            extensions=None,
            mandatory=True,
        ),
        rh_white=dict(
            extensions=None,
            mandatory=True,
        ),
        ribbon=dict(
            extensions=None,
            mandatory=True,
        ),
        subject_id=dict(
            argstr="%s",
            mandatory=True,
            position=-3,
            usedefault=True,
        ),
        subjects_dir=dict(),
        surface=dict(
            argstr="%s",
            position=-1,
        ),
        tabular_output=dict(
            argstr="-b",
        ),
        th3=dict(
            argstr="-th3",
            requires=["cortex_label"],
        ),
        thickness=dict(
            extensions=None,
            mandatory=True,
        ),
        transform=dict(
            extensions=None,
            mandatory=True,
        ),
        wm=dict(
            extensions=None,
            mandatory=True,
        ),
    )
    inputs = ParcellationStats.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_ParcellationStats_outputs():
    output_map = dict(
        out_color=dict(
            extensions=None,
        ),
        out_table=dict(
            extensions=None,
        ),
    )
    outputs = ParcellationStats.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# -*- coding: utf8 -*-
"""Autogenerated file - DO NOT EDIT
If you spot a bug, please report it on the mailing list and/or change the generator."""

from nipype.interfaces.base import (
    CommandLine,
    CommandLineInputSpec,
    SEMLikeCommandLine,
    TraitedSpec,
    File,
    Directory,
    traits,
    isdefined,
    InputMultiPath,
    OutputMultiPath,
)
import os


class SimpleRegionGrowingSegmentationInputSpec(CommandLineInputSpec):
    smoothingIterations = traits.Int(
        desc="Number of smoothing iterations", argstr="--smoothingIterations %d"
    )
    timestep = traits.Float(desc="Timestep for curvature flow", argstr="--timestep %f")
    iterations = traits.Int(
        desc="Number of iterations of region growing", argstr="--iterations %d"
    )
    multiplier = traits.Float(
        desc="Number of standard deviations to include in intensity model",
        argstr="--multiplier %f",
    )
    neighborhood = traits.Int(
        desc="The radius of the neighborhood over which to calculate intensity model",
        argstr="--neighborhood %d",
    )
    labelvalue = traits.Int(
        desc="The integer value (0-255) to use for the segmentation results. This will determine the color of the segmentation that will be generated by the Region growing algorithm",
        argstr="--labelvalue %d",
    )
    seed = InputMultiPath(
        traits.List(traits.Float(), minlen=3, maxlen=3),
        desc="Seed point(s) for region growing",
        argstr="--seed %s...",
    )
    inputVolume = File(
        position=-2, desc="Input volume to be filtered", exists=True, argstr="%s"
    )
    outputVolume = traits.Either(
        traits.Bool,
        File(),
        position=-1,
        hash_files=False,
        desc="Output filtered",
        argstr="%s",
    )


class SimpleRegionGrowingSegmentationOutputSpec(TraitedSpec):
    outputVolume = File(position=-1, desc="Output filtered", exists=True)


class SimpleRegionGrowingSegmentation(SEMLikeCommandLine):
    """title: Simple Region Growing Segmentation

category: Segmentation

description: A simple region growing segmentation algorithm based on intensity statistics. To create a list of fiducials (Seeds) for this algorithm, click on the tool bar icon of an arrow pointing to a starburst fiducial to enter the 'place a new object mode' and then use the fiducials module. This module uses the Slicer Command Line Interface (CLI) and the ITK filters CurvatureFlowImageFilter and ConfidenceConnectedImageFilter.

version: 0.1.0.$Revision: 19904 $(alpha)

documentation-url: http://www.slicer.org/slicerWiki/index.php/Documentation/4.1/Modules/SimpleRegionGrowingSegmentation

contributor: Jim Miller (GE)

acknowledgements: This command module was derived from Insight/Examples (copyright) Insight Software Consortium

"""

    input_spec = SimpleRegionGrowingSegmentationInputSpec
    output_spec = SimpleRegionGrowingSegmentationOutputSpec
    _cmd = "SimpleRegionGrowingSegmentation "
    _outputs_filenames = {"outputVolume": "outputVolume.nii"}



# -*- coding: utf-8 -*-
import os.path as op
import glob

from ... import logging
from ...utils.filemanip import simplify_list
from ..base import traits, File, Directory, TraitedSpec, OutputMultiPath
from ..freesurfer.base import FSCommand, FSTraitedSpec

iflogger = logging.getLogger("nipype.interface")


class WatershedBEMInputSpec(FSTraitedSpec):
    subject_id = traits.Str(
        argstr="--subject %s",
        mandatory=True,
        desc="Subject ID (must have a complete Freesurfer directory)",
    )
    subjects_dir = Directory(
        exists=True,
        mandatory=True,
        usedefault=True,
        desc="Path to Freesurfer subjects directory",
    )
    volume = traits.Enum(
        "T1",
        "aparc+aseg",
        "aseg",
        "brain",
        "orig",
        "brainmask",
        "ribbon",
        argstr="--volume %s",
        usedefault=True,
        desc='The volume from the "mri" directory to use (defaults to T1)',
    )
    overwrite = traits.Bool(
        True,
        usedefault=True,
        argstr="--overwrite",
        desc="Overwrites the existing files",
    )
    atlas_mode = traits.Bool(
        argstr="--atlas",
        desc="Use atlas mode for registration (default: no rigid alignment)",
    )


class WatershedBEMOutputSpec(TraitedSpec):
    mesh_files = OutputMultiPath(
        File(exists=True),
        desc=(
            "Paths to the output meshes (brain, inner "
            "skull, outer skull, outer skin)"
        ),
    )
    brain_surface = File(
        exists=True, loc="bem/watershed", desc="Brain surface (in Freesurfer format)"
    )
    inner_skull_surface = File(
        exists=True,
        loc="bem/watershed",
        desc="Inner skull surface (in Freesurfer format)",
    )
    outer_skull_surface = File(
        exists=True,
        loc="bem/watershed",
        desc="Outer skull surface (in Freesurfer format)",
    )
    outer_skin_surface = File(
        exists=True,
        loc="bem/watershed",
        desc="Outer skin surface (in Freesurfer format)",
    )
    fif_file = File(
        exists=True,
        loc="bem",
        altkey="fif",
        desc='"fif" format file for EEG processing in MNE',
    )
    cor_files = OutputMultiPath(
        File(exists=True),
        loc="bem/watershed/ws",
        altkey="COR",
        desc='"COR" format files',
    )


class WatershedBEM(FSCommand):
    """Uses mne_watershed_bem to get information from dicom directories

    Examples
    --------

    >>> from nipype.interfaces.mne import WatershedBEM
    >>> bem = WatershedBEM()
    >>> bem.inputs.subject_id = 'subj1'
    >>> bem.inputs.subjects_dir = '.'
    >>> bem.cmdline
    'mne watershed_bem --overwrite --subject subj1 --volume T1'
    >>> bem.run() 				# doctest: +SKIP

   """

    _cmd = "mne watershed_bem"
    input_spec = WatershedBEMInputSpec
    output_spec = WatershedBEMOutputSpec
    _additional_metadata = ["loc", "altkey"]

    def _get_files(self, path, key, dirval, altkey=None):
        globsuffix = "*"
        globprefix = "*"
        keydir = op.join(path, dirval)
        if altkey:
            key = altkey
        globpattern = op.join(keydir, "".join((globprefix, key, globsuffix)))
        return glob.glob(globpattern)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        subjects_dir = self.inputs.subjects_dir
        subject_path = op.join(subjects_dir, self.inputs.subject_id)
        output_traits = self._outputs()
        mesh_paths = []
        for k in list(outputs.keys()):
            if k != "mesh_files":
                val = self._get_files(
                    subject_path,
                    k,
                    output_traits.traits()[k].loc,
                    output_traits.traits()[k].altkey,
                )
                if val:
                    value_list = simplify_list(val)
                    if isinstance(value_list, list):
                        out_files = []
                        for value in value_list:
                            out_files.append(op.abspath(value))
                    elif isinstance(value_list, (str, bytes)):
                        out_files = op.abspath(value_list)
                    else:
                        raise TypeError
                    outputs[k] = out_files
                    if not k.rfind("surface") == -1:
                        mesh_paths.append(out_files)
        outputs["mesh_files"] = mesh_paths
        return outputs



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..model import MS_LDA


def test_MS_LDA_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        conform=dict(
            argstr="-conform",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        images=dict(
            argstr="%s",
            copyfile=False,
            mandatory=True,
            position=-1,
        ),
        label_file=dict(
            argstr="-label %s",
            extensions=None,
        ),
        lda_labels=dict(
            argstr="-lda %s",
            mandatory=True,
            sep=" ",
        ),
        mask_file=dict(
            argstr="-mask %s",
            extensions=None,
        ),
        shift=dict(
            argstr="-shift %d",
        ),
        subjects_dir=dict(),
        use_weights=dict(
            argstr="-W",
        ),
        vol_synth_file=dict(
            argstr="-synth %s",
            extensions=None,
            mandatory=True,
        ),
        weight_file=dict(
            argstr="-weight %s",
            extensions=None,
            mandatory=True,
        ),
    )
    inputs = MS_LDA.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_MS_LDA_outputs():
    output_map = dict(
        vol_synth_file=dict(
            extensions=None,
        ),
        weight_file=dict(
            extensions=None,
        ),
    )
    outputs = MS_LDA.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
""" Useful Functions for working with matlab"""

# Stdlib imports
import os
import re
import tempfile
import numpy as np

# Functions, classes and other top-level code


def fltcols(vals):
    """ Trivial little function to make 1xN float vector """
    return np.atleast_2d(np.array(vals, dtype=float))


def mlab_tempfile(dir=None):
    """Returns a temporary file-like object with valid matlab name.

    The file name is accessible as the .name attribute of the returned object.
    The caller is responsible for closing the returned object, at which time
    the underlying file gets deleted from the filesystem.

    Parameters
    ----------

      dir : str
        A path to use as the starting directory.  Note that this directory must
        already exist, it is NOT created if it doesn't (in that case, OSError
        is raised instead).

    Returns
    -------
      f : A file-like object.

    Examples
    --------

    >>> fn = mlab_tempfile()
    >>> import os
    >>> filename = os.path.basename(fn.name)
    >>> '-' not in filename
    True
    >>> fn.close()

    """

    valid_name = re.compile(r"^\w+$")

    # Make temp files until we get one whose name is a valid matlab identifier,
    # since matlab imposes that constraint.  Since the temp file routines may
    # return names that aren't valid matlab names, but we can't control that
    # directly, we just keep trying until we get a valid name.  To avoid an
    # infinite loop for some strange reason, we only try 100 times.
    for n in range(100):
        f = tempfile.NamedTemporaryFile(suffix=".m", prefix="tmp_matlab_", dir=dir)
        # Check the file name for matlab compilance
        fname = os.path.splitext(os.path.basename(f.name))[0]
        if valid_name.match(fname):
            break
        # Close the temp file we just made if its name is not valid; the
        # tempfile module then takes care of deleting the actual file on disk.
        f.close()
    else:
        raise ValueError("Could not make temp file after 100 tries")

    return f



import os
from .. import get_info
from ..info import get_nipype_gitversion
import pytest


def test_nipype_info():
    exception_not_raised = True
    try:
        get_info()
    except Exception:
        exception_not_raised = False
    assert exception_not_raised


@pytest.mark.skipif(
    not get_nipype_gitversion(),
    reason="not able to get version from get_nipype_gitversion",
)
def test_git_hash():
    # removing the first "g" from gitversion
    get_nipype_gitversion()[1:] == get_info()["commit_hash"]


def _check_no_et():
    import os
    from unittest.mock import patch

    et = os.getenv("NIPYPE_NO_ET") is None

    with patch.dict("os.environ", {"NIPYPE_NO_ET": "1"}):
        from nipype.interfaces.base import BaseInterface

        ver_data = BaseInterface._etelemetry_version_data

    if et and ver_data is None:
        raise ValueError(
            "etelemetry enabled and version data missing - double hits likely"
        )

    return et


def test_no_et_bare(tmp_path):
    from unittest.mock import patch
    from nipype.pipeline import engine as pe
    from nipype.interfaces import utility as niu
    from nipype.interfaces.base import BaseInterface

    et = os.getenv("NIPYPE_NO_ET") is None

    # Pytest doesn't trigger this, so let's pretend it's there
    with patch.object(BaseInterface, "_etelemetry_version_data", {}):

        # Direct function call - environment not set
        f = niu.Function(function=_check_no_et)
        res = f.run()
        assert res.outputs.out == et

        # Basic node - environment not set
        n = pe.Node(
            niu.Function(function=_check_no_et), name="n", base_dir=str(tmp_path)
        )
        res = n.run()
        assert res.outputs.out == et

        # Linear run - environment not set
        wf1 = pe.Workflow(name="wf1", base_dir=str(tmp_path))
        wf1.add_nodes([pe.Node(niu.Function(function=_check_no_et), name="n")])
        res = wf1.run()
        assert next(iter(res.nodes)).result.outputs.out == et


@pytest.mark.parametrize("plugin", ("MultiProc", "LegacyMultiProc"))
@pytest.mark.parametrize("run_without_submitting", (True, False))
def test_no_et_multiproc(tmp_path, plugin, run_without_submitting):
    from unittest.mock import patch
    from nipype.pipeline import engine as pe
    from nipype.interfaces import utility as niu
    from nipype.interfaces.base import BaseInterface

    et = os.getenv("NIPYPE_NO_ET") is None

    # Multiprocessing runs initialize new processes with NIPYPE_NO_ET
    # This does not apply to unsubmitted jobs, run by the main thread
    expectation = et if run_without_submitting else False

    # Pytest doesn't trigger this, so let's pretend it's there
    with patch.object(BaseInterface, "_etelemetry_version_data", {}):

        wf = pe.Workflow(name="wf2", base_dir=str(tmp_path))
        n = pe.Node(
            niu.Function(function=_check_no_et),
            run_without_submitting=run_without_submitting,
            name="n",
        )
        wf.add_nodes([n])
        res = wf.run(plugin=plugin, plugin_args={"n_procs": 1})
        assert next(iter(res.nodes)).result.outputs.out is expectation



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..minc import BBox


def test_BBox_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        format_minccrop=dict(
            argstr="-minccrop",
        ),
        format_mincresample=dict(
            argstr="-mincresample",
        ),
        format_mincreshape=dict(
            argstr="-mincreshape",
        ),
        input_file=dict(
            argstr="%s",
            extensions=None,
            mandatory=True,
            position=-2,
        ),
        one_line=dict(
            argstr="-one_line",
            xor=("one_line", "two_lines"),
        ),
        out_file=dict(
            argstr="> %s",
            extensions=None,
            genfile=True,
            position=-1,
        ),
        output_file=dict(
            extensions=None,
            hash_files=False,
            keep_extension=False,
            name_source=["input_file"],
            name_template="%s_bbox.txt",
            position=-1,
        ),
        threshold=dict(
            argstr="-threshold",
        ),
        two_lines=dict(
            argstr="-two_lines",
            xor=("one_line", "two_lines"),
        ),
    )
    inputs = BBox.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_BBox_outputs():
    output_map = dict(
        output_file=dict(
            extensions=None,
        ),
    )
    outputs = BBox.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
vtkbase provides some helpers to use VTK through the tvtk package (mayavi)

Code using tvtk should import it through this module
"""

import os
from .. import logging

iflogger = logging.getLogger("nipype.interface")

# Check that VTK can be imported and get version
_vtk_version = None
try:
    import vtk

    _vtk_version = (
        vtk.vtkVersion.GetVTKMajorVersion(),
        vtk.vtkVersion.GetVTKMinorVersion(),
    )
except ImportError:
    iflogger.warning("VTK was not found")

# Ensure that tvtk is loaded with the appropriate ETS_TOOLKIT env var
old_ets = os.getenv("ETS_TOOLKIT")
os.environ["ETS_TOOLKIT"] = "null"
_have_tvtk = False
try:
    from tvtk.api import tvtk

    _have_tvtk = True
except ImportError:
    iflogger.warning("tvtk wasn't found")
    tvtk = None
finally:
    if old_ets is not None:
        os.environ["ETS_TOOLKIT"] = old_ets
    else:
        del os.environ["ETS_TOOLKIT"]


def vtk_version():
    """ Get VTK version """
    global _vtk_version
    return _vtk_version


def no_vtk():
    """ Checks if VTK is installed and the python wrapper is functional """
    global _vtk_version
    return _vtk_version is None


def no_tvtk():
    """ Checks if tvtk was found """
    global _have_tvtk
    return not _have_tvtk


def vtk_old():
    """ Checks if VTK uses the old-style pipeline (VTK<6.0) """
    global _vtk_version
    if _vtk_version is None:
        raise RuntimeException("VTK is not correctly installed.")
    return _vtk_version[0] < 6


def configure_input_data(obj, data):
    """
    Configure the input data for vtk pipeline object obj.
    Copied from latest version of mayavi
    """
    if vtk_old():
        obj.input = data
    else:
        obj.set_input_data(data)


def vtk_output(obj):
    """ Configure the input data for vtk pipeline object obj."""
    if vtk_old():
        return obj.output
    return obj.get_output()



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..developer import MedicAlgorithmThresholdToBinaryMask


def test_MedicAlgorithmThresholdToBinaryMask_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        inLabel=dict(
            argstr="--inLabel %s",
            sep=";",
        ),
        inMaximum=dict(
            argstr="--inMaximum %f",
        ),
        inMinimum=dict(
            argstr="--inMinimum %f",
        ),
        inUse=dict(
            argstr="--inUse %s",
        ),
        null=dict(
            argstr="--null %s",
        ),
        outBinary=dict(
            argstr="--outBinary %s",
            sep=";",
        ),
        xDefaultMem=dict(
            argstr="-xDefaultMem %d",
        ),
        xMaxProcess=dict(
            argstr="-xMaxProcess %d",
            usedefault=True,
        ),
        xPrefExt=dict(
            argstr="--xPrefExt %s",
        ),
    )
    inputs = MedicAlgorithmThresholdToBinaryMask.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_MedicAlgorithmThresholdToBinaryMask_outputs():
    output_map = dict()
    outputs = MedicAlgorithmThresholdToBinaryMask.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
# -*- coding: utf-8 -*-

import os.path as op

from ..base import traits, TraitedSpec, File, Undefined, InputMultiObject
from .base import MRTrix3BaseInputSpec, MRTrix3Base


class FitTensorInputSpec(MRTrix3BaseInputSpec):
    in_file = File(
        exists=True,
        argstr="%s",
        mandatory=True,
        position=-2,
        desc="input diffusion weighted images",
    )
    out_file = File(
        "dti.mif",
        argstr="%s",
        mandatory=True,
        position=-1,
        usedefault=True,
        desc="the output diffusion tensor image",
    )

    # General options
    in_mask = File(
        exists=True,
        argstr="-mask %s",
        desc=(
            "only perform computation within the specified " "binary brain mask image"
        ),
    )
    method = traits.Enum(
        "nonlinear",
        "loglinear",
        "sech",
        "rician",
        argstr="-method %s",
        desc=("select method used to perform the fitting"),
    )
    reg_term = traits.Float(
        argstr="-regularisation %f",
        max_ver="0.3.13",
        desc=(
            "specify the strength of the regularisation term on the "
            "magnitude of the tensor elements (default = 5000). This "
            "only applies to the non-linear methods"
        ),
    )


class FitTensorOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="the output DTI file")


class FitTensor(MRTrix3Base):
    """
    Convert diffusion-weighted images to tensor images


    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> tsr = mrt.FitTensor()
    >>> tsr.inputs.in_file = 'dwi.mif'
    >>> tsr.inputs.in_mask = 'mask.nii.gz'
    >>> tsr.inputs.grad_fsl = ('bvecs', 'bvals')
    >>> tsr.cmdline                               # doctest: +ELLIPSIS
    'dwi2tensor -fslgrad bvecs bvals -mask mask.nii.gz dwi.mif dti.mif'
    >>> tsr.run()                                 # doctest: +SKIP
    """

    _cmd = "dwi2tensor"
    input_spec = FitTensorInputSpec
    output_spec = FitTensorOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = op.abspath(self.inputs.out_file)
        return outputs


class EstimateFODInputSpec(MRTrix3BaseInputSpec):
    algorithm = traits.Enum(
        "csd",
        "msmt_csd",
        argstr="%s",
        position=-8,
        mandatory=True,
        desc="FOD algorithm",
    )
    in_file = File(
        exists=True, argstr="%s", position=-7, mandatory=True, desc="input DWI image"
    )
    wm_txt = File(
        argstr="%s", position=-6, mandatory=True, desc="WM response text file"
    )
    wm_odf = File(
        "wm.mif",
        argstr="%s",
        position=-5,
        usedefault=True,
        mandatory=True,
        desc="output WM ODF",
    )
    gm_txt = File(argstr="%s", position=-4, desc="GM response text file")
    gm_odf = File(
        "gm.mif", usedefault=True, argstr="%s", position=-3, desc="output GM ODF"
    )
    csf_txt = File(argstr="%s", position=-2, desc="CSF response text file")
    csf_odf = File(
        "csf.mif", usedefault=True, argstr="%s", position=-1, desc="output CSF ODF"
    )
    mask_file = File(exists=True, argstr="-mask %s", desc="mask image")

    # DW Shell selection options
    shell = traits.List(
        traits.Float,
        sep=",",
        argstr="-shell %s",
        desc="specify one or more dw gradient shells",
    )
    max_sh = InputMultiObject(
        traits.Int,
        value=[8],
        usedefault=True,
        argstr="-lmax %s",
        sep=",",
        desc=(
            "maximum harmonic degree of response function - single value for single-shell response, list for multi-shell response"
        ),
    )
    in_dirs = File(
        exists=True,
        argstr="-directions %s",
        desc=(
            "specify the directions over which to apply the non-negativity "
            "constraint (by default, the built-in 300 direction set is "
            "used). These should be supplied as a text file containing the "
            "[ az el ] pairs for the directions."
        ),
    )


class EstimateFODOutputSpec(TraitedSpec):
    wm_odf = File(argstr="%s", desc="output WM ODF")
    gm_odf = File(argstr="%s", desc="output GM ODF")
    csf_odf = File(argstr="%s", desc="output CSF ODF")


class EstimateFOD(MRTrix3Base):
    """
    Estimate fibre orientation distributions from diffusion data using spherical deconvolution

    .. warning::

       The CSD algorithm does not work as intended, but fixing it in this interface could break
       existing workflows. This interface has been superseded by
       :py:class:`.ConstrainedSphericalDecomposition`.

    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> fod = mrt.EstimateFOD()
    >>> fod.inputs.algorithm = 'msmt_csd'
    >>> fod.inputs.in_file = 'dwi.mif'
    >>> fod.inputs.wm_txt = 'wm.txt'
    >>> fod.inputs.grad_fsl = ('bvecs', 'bvals')
    >>> fod.cmdline
    'dwi2fod -fslgrad bvecs bvals -lmax 8 msmt_csd dwi.mif wm.txt wm.mif gm.mif csf.mif'
    >>> fod.run()  # doctest: +SKIP
    """

    _cmd = "dwi2fod"
    input_spec = EstimateFODInputSpec
    output_spec = EstimateFODOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["wm_odf"] = op.abspath(self.inputs.wm_odf)
        if self.inputs.gm_odf != Undefined:
            outputs["gm_odf"] = op.abspath(self.inputs.gm_odf)
        if self.inputs.csf_odf != Undefined:
            outputs["csf_odf"] = op.abspath(self.inputs.csf_odf)
        return outputs


class ConstrainedSphericalDeconvolutionInputSpec(EstimateFODInputSpec):
    gm_odf = File(argstr="%s", position=-3, desc="output GM ODF")
    csf_odf = File(argstr="%s", position=-1, desc="output CSF ODF")
    max_sh = InputMultiObject(
        traits.Int,
        argstr="-lmax %s",
        sep=",",
        desc=(
            "maximum harmonic degree of response function - single value for single-shell response, list for multi-shell response"
        ),
    )


class ConstrainedSphericalDeconvolution(EstimateFOD):
    """
    Estimate fibre orientation distributions from diffusion data using spherical deconvolution

    This interface supersedes :py:class:`.EstimateFOD`.
    The old interface has contained a bug when using the CSD algorithm as opposed to the MSMT CSD
    algorithm, but fixing it could potentially break existing workflows. The new interface works
    the same, but does not populate the following inputs by default:

    * ``gm_odf``
    * ``csf_odf``
    * ``max_sh``

    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> fod = mrt.ConstrainedSphericalDeconvolution()
    >>> fod.inputs.algorithm = 'csd'
    >>> fod.inputs.in_file = 'dwi.mif'
    >>> fod.inputs.wm_txt = 'wm.txt'
    >>> fod.inputs.grad_fsl = ('bvecs', 'bvals')
    >>> fod.cmdline
    'dwi2fod -fslgrad bvecs bvals csd dwi.mif wm.txt wm.mif'
    >>> fod.run()  # doctest: +SKIP
    """

    input_spec = ConstrainedSphericalDeconvolutionInputSpec



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import os
import copy
import simplejson
import glob
import os.path as op
from subprocess import Popen
import hashlib
from collections import namedtuple

import pytest
import nipype
import nipype.interfaces.io as nio
from nipype.interfaces.base.traits_extension import isdefined
from nipype.interfaces.base import Undefined, TraitError
from nipype.utils.filemanip import dist_is_editable

# Check for boto
noboto = False
try:
    import boto
    from boto.s3.connection import S3Connection, OrdinaryCallingFormat
except ImportError:
    noboto = True

# Check for boto3
noboto3 = False
try:
    import boto3
    from botocore.utils import fix_s3_host
except ImportError:
    noboto3 = True

# Check for paramiko
try:
    import paramiko

    no_paramiko = False

    # Check for localhost SSH Server
    # FIXME: Tests requiring this are never run on CI
    try:
        proxy = None
        client = paramiko.SSHClient()
        client.load_system_host_keys()
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        client.connect("127.0.0.1", username=os.getenv("USER"), sock=proxy, timeout=10)

        no_local_ssh = False

    except (
        paramiko.SSHException,
        paramiko.ssh_exception.NoValidConnectionsError,
        OSError,
    ):
        no_local_ssh = True

except ImportError:
    no_paramiko = True
    no_local_ssh = True

# Check for fakes3
from subprocess import check_call, CalledProcessError

try:
    ret_code = check_call(["which", "fakes3"], stdout=open(os.devnull, "wb"))
    fakes3 = ret_code == 0
except CalledProcessError:
    fakes3 = False

# check for bids
have_pybids = True
try:
    import bids

    filepath = os.path.realpath(os.path.dirname(bids.__file__))
    datadir = os.path.realpath(os.path.join(filepath, "tests/data/"))
except ImportError:
    have_pybids = False


def test_datagrabber():
    dg = nio.DataGrabber()
    assert dg.inputs.template == Undefined
    assert dg.inputs.base_directory == Undefined
    assert dg.inputs.template_args == {"outfiles": []}


@pytest.mark.skipif(noboto, reason="boto library is not available")
def test_s3datagrabber():
    dg = nio.S3DataGrabber()
    assert dg.inputs.template == Undefined
    assert dg.inputs.local_directory == Undefined
    assert dg.inputs.template_args == {"outfiles": []}


templates1 = {
    "model": "interfaces/{package}/model.py",
    "preprocess": "interfaces/{package}/pre*.py",
}
templates2 = {"converter": "interfaces/dcm{to!s}nii.py"}
templates3 = {"model": "interfaces/{package.name}/model.py"}


@pytest.mark.parametrize(
    "SF_args, inputs_att, expected",
    [
        (
            {"templates": templates1},
            {"package": "fsl"},
            {
                "infields": ["package"],
                "outfields": ["model", "preprocess"],
                "run_output": {
                    "model": op.join(
                        op.dirname(nipype.__file__), "interfaces/fsl/model.py"
                    ),
                    "preprocess": op.join(
                        op.dirname(nipype.__file__), "interfaces/fsl/preprocess.py"
                    ),
                },
                "node_output": ["model", "preprocess"],
            },
        ),
        (
            {"templates": templates1, "force_lists": True},
            {"package": "spm"},
            {
                "infields": ["package"],
                "outfields": ["model", "preprocess"],
                "run_output": {
                    "model": [
                        op.join(op.dirname(nipype.__file__), "interfaces/spm/model.py")
                    ],
                    "preprocess": [
                        op.join(
                            op.dirname(nipype.__file__), "interfaces/spm/preprocess.py"
                        )
                    ],
                },
                "node_output": ["model", "preprocess"],
            },
        ),
        (
            {"templates": templates1},
            {"package": "fsl", "force_lists": ["model"]},
            {
                "infields": ["package"],
                "outfields": ["model", "preprocess"],
                "run_output": {
                    "model": [
                        op.join(op.dirname(nipype.__file__), "interfaces/fsl/model.py")
                    ],
                    "preprocess": op.join(
                        op.dirname(nipype.__file__), "interfaces/fsl/preprocess.py"
                    ),
                },
                "node_output": ["model", "preprocess"],
            },
        ),
        (
            {"templates": templates2},
            {"to": 2},
            {
                "infields": ["to"],
                "outfields": ["converter"],
                "run_output": {
                    "converter": op.join(
                        op.dirname(nipype.__file__), "interfaces/dcm2nii.py"
                    )
                },
                "node_output": ["converter"],
            },
        ),
        (
            {"templates": templates3},
            {"package": namedtuple("package", ["name"])("fsl")},
            {
                "infields": ["package"],
                "outfields": ["model"],
                "run_output": {
                    "model": op.join(
                        op.dirname(nipype.__file__), "interfaces/fsl/model.py"
                    )
                },
                "node_output": ["model"],
            },
        ),
    ],
)
def test_selectfiles(tmpdir, SF_args, inputs_att, expected):
    tmpdir.chdir()
    base_dir = op.dirname(nipype.__file__)
    dg = nio.SelectFiles(base_directory=base_dir, **SF_args)
    for key, val in inputs_att.items():
        setattr(dg.inputs, key, val)

    assert dg._infields == expected["infields"]
    assert sorted(dg._outfields) == expected["outfields"]
    assert sorted(dg._outputs().get()) == expected["node_output"]

    res = dg.run()
    for key, val in expected["run_output"].items():
        assert getattr(res.outputs, key) == val


def test_selectfiles_valueerror():
    """Test ValueError when force_lists has field that isn't in template."""
    base_dir = op.dirname(nipype.__file__)
    templates = {
        "model": "interfaces/{package}/model.py",
        "preprocess": "interfaces/{package}/pre*.py",
    }
    force_lists = ["model", "preprocess", "registration"]
    sf = nio.SelectFiles(templates, base_directory=base_dir, force_lists=force_lists)
    with pytest.raises(ValueError):
        sf.run()


@pytest.mark.skipif(noboto, reason="boto library is not available")
def test_s3datagrabber_communication(tmpdir):
    dg = nio.S3DataGrabber(
        infields=["subj_id", "run_num"], outfields=["func", "struct"]
    )
    dg.inputs.anon = True
    dg.inputs.bucket = "openfmri"
    dg.inputs.bucket_path = "ds001/"
    dg.inputs.local_directory = tmpdir.strpath
    dg.inputs.sort_filelist = True
    dg.inputs.template = "*"
    dg.inputs.field_template = dict(
        func="%s/BOLD/task001_%s/bold.nii.gz",
        struct="%s/anatomy/highres001_brain.nii.gz",
    )
    dg.inputs.subj_id = ["sub001", "sub002"]
    dg.inputs.run_num = ["run001", "run003"]
    dg.inputs.template_args = dict(func=[["subj_id", "run_num"]], struct=[["subj_id"]])
    res = dg.run()
    func_outfiles = res.outputs.func
    struct_outfiles = res.outputs.struct

    # check for all files
    assert (
        os.path.join(
            dg.inputs.local_directory, "/sub001/BOLD/task001_run001/bold.nii.gz"
        )
        in func_outfiles[0]
    )
    assert os.path.exists(func_outfiles[0])
    assert (
        os.path.join(
            dg.inputs.local_directory, "/sub001/anatomy/highres001_brain.nii.gz"
        )
        in struct_outfiles[0]
    )
    assert os.path.exists(struct_outfiles[0])
    assert (
        os.path.join(
            dg.inputs.local_directory, "/sub002/BOLD/task001_run003/bold.nii.gz"
        )
        in func_outfiles[1]
    )
    assert os.path.exists(func_outfiles[1])
    assert (
        os.path.join(
            dg.inputs.local_directory, "/sub002/anatomy/highres001_brain.nii.gz"
        )
        in struct_outfiles[1]
    )
    assert os.path.exists(struct_outfiles[1])


def test_datagrabber_order(tmpdir):
    for file_name in [
        "sub002_L1_R1.q",
        "sub002_L1_R2.q",
        "sub002_L2_R1.q",
        "sub002_L2_R2.qd",
        "sub002_L3_R10.q",
        "sub002_L3_R2.q",
    ]:
        tmpdir.join(file_name).open("a").close()

    dg = nio.DataGrabber(infields=["sid"])
    dg.inputs.base_directory = tmpdir.strpath
    dg.inputs.template = "%s_L%d_R*.q*"
    dg.inputs.template_args = {"outfiles": [["sid", 1], ["sid", 2], ["sid", 3]]}
    dg.inputs.sid = "sub002"
    dg.inputs.sort_filelist = True
    res = dg.run()
    outfiles = res.outputs.outfiles

    assert "sub002_L1_R1" in outfiles[0][0]
    assert "sub002_L1_R2" in outfiles[0][1]
    assert "sub002_L2_R1" in outfiles[1][0]
    assert "sub002_L2_R2" in outfiles[1][1]
    assert "sub002_L3_R2" in outfiles[2][0]
    assert "sub002_L3_R10" in outfiles[2][1]


def test_datasink():
    ds = nio.DataSink()
    assert ds.inputs.parameterization
    assert ds.inputs.base_directory == Undefined
    assert ds.inputs.strip_dir == Undefined
    assert ds.inputs._outputs == {}

    ds = nio.DataSink(base_directory="foo")
    assert ds.inputs.base_directory == "foo"

    ds = nio.DataSink(infields=["test"])
    assert "test" in ds.inputs.copyable_trait_names()


# Make dummy input file
@pytest.fixture(scope="module")
def dummy_input(request, tmpdir_factory):
    """
    Function to create a dummy file
    """
    # Init variables

    input_path = tmpdir_factory.mktemp("input_data").join("datasink_test_s3.txt")

    # Create input file
    input_path.write_binary(b"ABCD1234")

    # Return path
    return str(input_path)


# Test datasink writes to s3 properly
@pytest.mark.skipif(
    noboto3 or not fakes3, reason="boto3 or fakes3 library is not available"
)
def test_datasink_to_s3(dummy_input, tmpdir):
    """
    This function tests to see if the S3 functionality of a DataSink
    works properly
    """
    # Init variables
    ds = nio.DataSink()
    bucket_name = "test"
    container = "outputs"
    attr_folder = "text_file"
    output_dir = "s3://" + bucket_name
    # Local temporary filepaths for testing
    fakes3_dir = tmpdir.strpath
    input_path = dummy_input

    # Start up fake-S3 server
    proc = Popen(
        ["fakes3", "-r", fakes3_dir, "-p", "4567"], stdout=open(os.devnull, "wb")
    )

    # Init boto3 s3 resource to talk with fakes3
    resource = boto3.resource(
        aws_access_key_id="mykey",
        aws_secret_access_key="mysecret",
        service_name="s3",
        endpoint_url="http://127.0.0.1:4567",
        use_ssl=False,
    )
    resource.meta.client.meta.events.unregister("before-sign.s3", fix_s3_host)

    # Create bucket
    bucket = resource.create_bucket(Bucket=bucket_name)

    # Prep datasink
    ds.inputs.base_directory = output_dir
    ds.inputs.container = container
    ds.inputs.bucket = bucket
    setattr(ds.inputs, attr_folder, input_path)

    # Run datasink
    ds.run()

    # Get MD5sums and compare
    key = "/".join([container, attr_folder, os.path.basename(input_path)])
    obj = bucket.Object(key=key)
    dst_md5 = obj.e_tag.replace('"', "")
    src_md5 = hashlib.md5(open(input_path, "rb").read()).hexdigest()

    # Kill fakes3
    proc.kill()

    # Make sure md5sums match
    assert src_md5 == dst_md5


# Test AWS creds read from env vars
@pytest.mark.skipif(
    noboto3 or not fakes3, reason="boto3 or fakes3 library is not available"
)
def test_aws_keys_from_env():
    """
    Function to ensure the DataSink can successfully read in AWS
    credentials from the environment variables
    """

    # Init variables
    ds = nio.DataSink()
    aws_access_key_id = "ABCDACCESS"
    aws_secret_access_key = "DEFGSECRET"

    # Set env vars
    os.environ["AWS_ACCESS_KEY_ID"] = aws_access_key_id
    os.environ["AWS_SECRET_ACCESS_KEY"] = aws_secret_access_key

    # Call function to return creds
    access_key_test, secret_key_test = ds._return_aws_keys()

    # Assert match
    assert aws_access_key_id == access_key_test
    assert aws_secret_access_key == secret_key_test


# Test the local copy attribute
def test_datasink_localcopy(dummy_input, tmpdir):
    """
    Function to validate DataSink will make local copy via local_copy
    attribute
    """

    # Init variables
    local_dir = tmpdir.strpath
    container = "outputs"
    attr_folder = "text_file"

    # Make dummy input file and datasink
    input_path = dummy_input

    ds = nio.DataSink()

    # Set up datasink
    ds.inputs.container = container
    ds.inputs.local_copy = local_dir

    setattr(ds.inputs, attr_folder, input_path)

    # Expected local copy path
    local_copy = os.path.join(
        local_dir, container, attr_folder, os.path.basename(input_path)
    )

    # Run the datasink
    ds.run()

    # Check md5sums of both
    src_md5 = hashlib.md5(open(input_path, "rb").read()).hexdigest()
    dst_md5 = hashlib.md5(open(local_copy, "rb").read()).hexdigest()

    # Perform test
    assert src_md5 == dst_md5


def test_datasink_substitutions(tmpdir):
    indir = tmpdir.mkdir("-Tmp-nipype_ds_subs_in")
    outdir = tmpdir.mkdir("-Tmp-nipype_ds_subs_out")
    files = []
    for n in ["ababab.n", "xabababyz.n"]:
        f = str(indir.join(n))
        files.append(f)
        open(f, "w")
    ds = nio.DataSink(
        parameterization=False,
        base_directory=str(outdir),
        substitutions=[("ababab", "ABABAB")],
        # end archoring ($) is used to assure operation on the filename
        # instead of possible temporary directories names matches
        # Patterns should be more comprehendable in the real-world usage
        # cases since paths would be quite more sensible
        regexp_substitutions=[
            (r"xABABAB(\w*)\.n$", r"a-\1-b.n"),
            ("(.*%s)[-a]([^%s]*)$" % ((os.path.sep,) * 2), r"\1!\2"),
        ],
    )
    setattr(ds.inputs, "@outdir", files)
    ds.run()
    assert sorted(
        [os.path.basename(x) for x in glob.glob(os.path.join(str(outdir), "*"))]
    ) == [
        "!-yz-b.n",
        "ABABAB.n",
    ]  # so we got re used 2nd and both patterns


@pytest.fixture()
def _temp_analyze_files(tmpdir):
    """Generate temporary analyze file pair."""
    img_dir = tmpdir.mkdir("img")
    orig_img = img_dir.join("orig.img")
    orig_hdr = img_dir.join("orig.hdr")
    orig_img.open("w")
    orig_hdr.open("w")
    return orig_img.strpath, orig_hdr.strpath


def test_datasink_copydir_1(_temp_analyze_files, tmpdir):
    orig_img, orig_hdr = _temp_analyze_files
    outdir = tmpdir
    pth, fname = os.path.split(orig_img)
    ds = nio.DataSink(
        base_directory=outdir.mkdir("basedir").strpath, parameterization=False
    )
    setattr(ds.inputs, "@outdir", pth)
    ds.run()
    sep = os.path.sep
    assert tmpdir.join("basedir", pth.split(sep)[-1], fname).check()


def test_datasink_copydir_2(_temp_analyze_files, tmpdir):
    orig_img, orig_hdr = _temp_analyze_files
    pth, fname = os.path.split(orig_img)
    ds = nio.DataSink(
        base_directory=tmpdir.mkdir("basedir").strpath, parameterization=False
    )
    ds.inputs.remove_dest_dir = True
    setattr(ds.inputs, "outdir", pth)
    ds.run()
    sep = os.path.sep
    assert not tmpdir.join("basedir", pth.split(sep)[-1], fname).check()
    assert tmpdir.join("basedir", "outdir", pth.split(sep)[-1], fname).check()


def test_datafinder_depth(tmpdir):
    outdir = tmpdir.strpath
    os.makedirs(os.path.join(outdir, "0", "1", "2", "3"))

    df = nio.DataFinder()
    df.inputs.root_paths = os.path.join(outdir, "0")
    for min_depth in range(4):
        for max_depth in range(min_depth, 4):
            df.inputs.min_depth = min_depth
            df.inputs.max_depth = max_depth
            result = df.run()
            expected = ["{}".format(x) for x in range(min_depth, max_depth + 1)]
            for path, exp_fname in zip(result.outputs.out_paths, expected):
                _, fname = os.path.split(path)
                assert fname == exp_fname


def test_datafinder_unpack(tmpdir):
    outdir = tmpdir.strpath
    single_res = os.path.join(outdir, "findme.txt")
    open(single_res, "a").close()
    open(os.path.join(outdir, "dontfindme"), "a").close()

    df = nio.DataFinder()
    df.inputs.root_paths = outdir
    df.inputs.match_regex = r".+/(?P<basename>.+)\.txt"
    df.inputs.unpack_single = True
    result = df.run()
    print(result.outputs.out_paths)
    assert result.outputs.out_paths == single_res


def test_freesurfersource():
    fss = nio.FreeSurferSource()
    assert fss.inputs.hemi == "both"
    assert fss.inputs.subject_id == Undefined
    assert fss.inputs.subjects_dir == Undefined


def test_freesurfersource_incorrectdir():
    fss = nio.FreeSurferSource()
    with pytest.raises(TraitError) as err:
        fss.inputs.subjects_dir = "path/to/no/existing/directory"


def test_jsonsink_input():

    ds = nio.JSONFileSink()
    assert ds.inputs._outputs == {}

    ds = nio.JSONFileSink(in_dict={"foo": "var"})
    assert ds.inputs.in_dict == {"foo": "var"}

    ds = nio.JSONFileSink(infields=["test"])
    assert "test" in ds.inputs.copyable_trait_names()


@pytest.mark.parametrize(
    "inputs_attributes",
    [{"new_entry": "someValue"}, {"new_entry": "someValue", "test": "testInfields"}],
)
def test_jsonsink(tmpdir, inputs_attributes):
    tmpdir.chdir()
    js = nio.JSONFileSink(infields=["test"], in_dict={"foo": "var"})
    setattr(js.inputs, "contrasts.alt", "someNestedValue")
    expected_data = {"contrasts": {"alt": "someNestedValue"}, "foo": "var"}
    for key, val in inputs_attributes.items():
        setattr(js.inputs, key, val)
        expected_data[key] = val

    res = js.run()
    with open(res.outputs.out_file, "r") as f:
        data = simplejson.load(f)

    assert data == expected_data


# There are three reasons these tests will be skipped:
@pytest.mark.skipif(not have_pybids, reason="Pybids is not installed")
@pytest.mark.skipif(
    not dist_is_editable("pybids"), reason="Pybids is not installed in editable mode"
)
def test_bids_grabber(tmpdir):
    tmpdir.chdir()
    bg = nio.BIDSDataGrabber()
    bg.inputs.base_dir = os.path.join(datadir, "ds005")
    bg.inputs.subject = "01"
    results = bg.run()
    assert "sub-01_T1w.nii.gz" in map(os.path.basename, results.outputs.T1w)
    assert "sub-01_task-mixedgamblestask_run-01_bold.nii.gz" in map(
        os.path.basename, results.outputs.bold
    )


@pytest.mark.skipif(not have_pybids, reason="Pybids is not installed")
@pytest.mark.skipif(
    not dist_is_editable("pybids"), reason="Pybids is not installed in editable mode"
)
def test_bids_fields(tmpdir):
    tmpdir.chdir()
    bg = nio.BIDSDataGrabber(infields=["subject"], outfields=["dwi"])
    bg.inputs.base_dir = os.path.join(datadir, "ds005")
    bg.inputs.subject = "01"
    bg.inputs.output_query["dwi"] = dict(datatype="dwi")
    results = bg.run()
    assert "sub-01_dwi.nii.gz" in map(os.path.basename, results.outputs.dwi)


@pytest.mark.skipif(not have_pybids, reason="Pybids is not installed")
@pytest.mark.skipif(
    not dist_is_editable("pybids"), reason="Pybids is not installed in editable mode"
)
def test_bids_infields_outfields(tmpdir):
    tmpdir.chdir()
    infields = ["infield1", "infield2"]
    outfields = ["outfield1", "outfield2"]
    bg = nio.BIDSDataGrabber(infields=infields)
    for outfield in outfields:
        bg.inputs.output_query[outfield] = {"key": "value"}

    for infield in infields:
        assert infield in bg.inputs.traits()
        assert not (isdefined(bg.inputs.get()[infield]))

    for outfield in outfields:
        assert outfield in bg._outputs().traits()

    # now try without defining outfields
    bg = nio.BIDSDataGrabber()
    for outfield in ["T1w", "bold"]:
        assert outfield in bg._outputs().traits()


@pytest.mark.skipif(no_paramiko, reason="paramiko library is not available")
@pytest.mark.skipif(no_local_ssh, reason="SSH Server is not running")
def test_SSHDataGrabber(tmpdir):
    """Test SSHDataGrabber by connecting to localhost and collecting some data.
    """
    old_cwd = tmpdir.chdir()

    source_dir = tmpdir.mkdir("source")
    source_hdr = source_dir.join("somedata.hdr")
    source_dat = source_dir.join("somedata.img")
    source_hdr.ensure()  # create
    source_dat.ensure()  # create

    # ssh client that connects to localhost, current user, regardless of
    # ~/.ssh/config
    def _mock_get_ssh_client(self):
        proxy = None
        client = paramiko.SSHClient()
        client.load_system_host_keys()
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        client.connect("127.0.0.1", username=os.getenv("USER"), sock=proxy, timeout=10)
        return client

    MockSSHDataGrabber = copy.copy(nio.SSHDataGrabber)
    MockSSHDataGrabber._get_ssh_client = _mock_get_ssh_client

    # grabber to get files from source_dir matching test.hdr
    ssh_grabber = MockSSHDataGrabber(infields=["test"], outfields=["test_file"])
    ssh_grabber.inputs.base_directory = str(source_dir)
    ssh_grabber.inputs.hostname = "127.0.0.1"
    ssh_grabber.inputs.field_template = dict(test_file="%s.hdr")
    ssh_grabber.inputs.template = ""
    ssh_grabber.inputs.template_args = dict(test_file=[["test"]])
    ssh_grabber.inputs.test = "somedata"
    ssh_grabber.inputs.sort_filelist = True

    runtime = ssh_grabber.run()

    # did we successfully get the header?
    assert runtime.outputs.test_file == str(tmpdir.join(source_hdr.basename))
    # did we successfully get the data?
    assert (
        tmpdir.join(source_hdr.basename)  # header file
        .new(ext=".img")  # data file
        .check(file=True, exists=True)
    )  # exists?

    old_cwd.chdir()


def test_ExportFile(tmp_path):
    testin = tmp_path / "in.txt"
    testin.write_text("test string")
    i = nio.ExportFile()
    i.inputs.in_file = str(testin)
    i.inputs.out_file = str(tmp_path / "out.tsv")
    i.inputs.check_extension = True
    with pytest.raises(RuntimeError):
        i.run()
    i.inputs.check_extension = False
    i.run()
    assert (tmp_path / "out.tsv").read_text() == "test string"
    i.inputs.out_file = str(tmp_path / "out.txt")
    i.inputs.check_extension = True
    i.run()
    assert (tmp_path / "out.txt").read_text() == "test string"
    with pytest.raises(FileExistsError):
        i.run()
    i.inputs.clobber = True
    i.run()
    assert (tmp_path / "out.txt").read_text() == "test string"



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""

Miscellaneous tools to support Interface functionality
......................................................

"""
import os
from copy import deepcopy
from textwrap import wrap
import re

from ... import logging
from ...utils.misc import is_container
from ...utils.filemanip import md5, hash_infile

iflogger = logging.getLogger("nipype.interface")

HELP_LINEWIDTH = 70


class NipypeInterfaceError(Exception):
    """Custom error for interfaces"""

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return "{}".format(self.value)


class Bunch(object):
    """
    Dictionary-like class that provides attribute-style access to its items.

    A ``Bunch`` is a simple container that stores its items as class
    attributes [1]_. Internally all items are stored in a dictionary and
    the class exposes several of the dictionary methods.

    Examples
    --------
    >>> from nipype.interfaces.base import Bunch
    >>> inputs = Bunch(infile='subj.nii', fwhm=6.0, register_to_mean=True)
    >>> inputs
    Bunch(fwhm=6.0, infile='subj.nii', register_to_mean=True)
    >>> inputs.register_to_mean = False
    >>> inputs
    Bunch(fwhm=6.0, infile='subj.nii', register_to_mean=False)

    References
    ----------
    .. [1] A. Martelli, D. Hudgeon, "Collecting a Bunch of Named
           Items", Python Cookbook, 2nd Ed, Chapter 4.18, 2005.

    """

    def __init__(self, *args, **kwargs):
        self.__dict__.update(*args, **kwargs)

    def update(self, *args, **kwargs):
        """update existing attribute, or create new attribute

        Note: update is very much like HasTraits.set"""
        self.__dict__.update(*args, **kwargs)

    def items(self):
        """iterates over bunch attributes as key, value pairs"""
        return list(self.__dict__.items())

    def iteritems(self):
        """iterates over bunch attributes as key, value pairs"""
        iflogger.warning("iteritems is deprecated, use items instead")
        return list(self.items())

    def get(self, *args):
        """Support dictionary get() functionality
        """
        return self.__dict__.get(*args)

    def set(self, **kwargs):
        """Support dictionary get() functionality
        """
        return self.__dict__.update(**kwargs)

    def dictcopy(self):
        """returns a deep copy of existing Bunch as a dictionary"""
        return deepcopy(self.__dict__)

    def __repr__(self):
        """representation of the sorted Bunch as a string

        Currently, this string representation of the `inputs` Bunch of
        interfaces is hashed to determine if the process' dirty-bit
        needs setting or not. Till that mechanism changes, only alter
        this after careful consideration.
        """
        outstr = ["Bunch("]
        first = True
        for k, v in sorted(self.items()):
            if not first:
                outstr.append(", ")
            if isinstance(v, dict):
                pairs = []
                for key, value in sorted(v.items()):
                    pairs.append("'%s': %s" % (key, value))
                v = "{" + ", ".join(pairs) + "}"
                outstr.append("%s=%s" % (k, v))
            else:
                outstr.append("%s=%r" % (k, v))
            first = False
        outstr.append(")")
        return "".join(outstr)

    def _get_bunch_hash(self):
        """Return a dictionary of our items with hashes for each file.

        Searches through dictionary items and if an item is a file, it
        calculates the md5 hash of the file contents and stores the
        file name and hash value as the new key value.

        However, the overall bunch hash is calculated only on the hash
        value of a file. The path and name of the file are not used in
        the overall hash calculation.

        Returns
        -------
        dict_withhash : dict
            Copy of our dictionary with the new file hashes included
            with each file.
        hashvalue : str
            The md5 hash value of the `dict_withhash`

        """

        infile_list = []
        for key, val in list(self.items()):
            if is_container(val):
                # XXX - SG this probably doesn't catch numpy arrays
                # containing embedded file names either.
                if isinstance(val, dict):
                    # XXX - SG should traverse dicts, but ignoring for now
                    item = None
                else:
                    if len(val) == 0:
                        raise AttributeError("%s attribute is empty" % key)
                    item = val[0]
            else:
                item = val
            try:
                if isinstance(item, str) and os.path.isfile(item):
                    infile_list.append(key)
            except TypeError:
                # `item` is not a file or string.
                continue
        dict_withhash = self.dictcopy()
        dict_nofilename = self.dictcopy()
        for item in infile_list:
            dict_withhash[item] = _hash_bunch_dict(dict_withhash, item)
            dict_nofilename[item] = [val[1] for val in dict_withhash[item]]
        # Sort the items of the dictionary, before hashing the string
        # representation so we get a predictable order of the
        # dictionary.
        sorted_dict = str(sorted(dict_nofilename.items()))
        return dict_withhash, md5(sorted_dict.encode()).hexdigest()

    def _repr_pretty_(self, p, cycle):
        """Support for the pretty module from ipython.externals"""
        if cycle:
            p.text("Bunch(...)")
        else:
            p.begin_group(6, "Bunch(")
            first = True
            for k, v in sorted(self.items()):
                if not first:
                    p.text(",")
                    p.breakable()
                p.text(k + "=")
                p.pretty(v)
                first = False
            p.end_group(6, ")")


def _hash_bunch_dict(adict, key):
    """Inject file hashes into adict[key]"""
    stuff = adict[key]
    if not is_container(stuff):
        stuff = [stuff]
    return [(afile, hash_infile(afile)) for afile in stuff]


class InterfaceResult(object):
    """Object that contains the results of running a particular Interface.

    Attributes
    ----------
    version : version of this Interface result object (a readonly property)
    interface : class type
        A copy of the `Interface` class that was run to generate this result.
    inputs :  a traits free representation of the inputs
    outputs : Bunch
        An `Interface` specific Bunch that contains all possible files
        that are generated by the interface.  The `outputs` are used
        as the `inputs` to another node when interfaces are used in
        the pipeline.
    runtime : Bunch

        Contains attributes that describe the runtime environment when
        the `Interface` was run.  Contains the attributes:

        * cmdline : The command line string that was executed
        * cwd : The directory the ``cmdline`` was executed in.
        * stdout : The output of running the ``cmdline``.
        * stderr : Any error messages output from running ``cmdline``.
        * returncode : The code returned from running the ``cmdline``.

    """

    def __init__(self, interface, runtime, inputs=None, outputs=None, provenance=None):
        self._version = 2.0
        self.interface = interface
        self.runtime = runtime
        self.inputs = inputs
        self.outputs = outputs
        self.provenance = provenance

    @property
    def version(self):
        return self._version


def format_help(cls):
    """
    Prints help text of a Nipype interface

    >>> from nipype.interfaces.afni import GCOR
    >>> GCOR.help()  # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
    Wraps the executable command ``@compute_gcor``.
    <BLANKLINE>
    Computes the average correlation between every voxel
    and ever other voxel, over any give mask.
    <BLANKLINE>
    <BLANKLINE>
    For complete details, ...

    """
    from ...utils.misc import trim

    docstring = []
    cmd = getattr(cls, "_cmd", None)
    if cmd:
        docstring += ["Wraps the executable command ``%s``." % cmd, ""]

    if cls.__doc__:
        docstring += trim(cls.__doc__).split("\n") + [""]

    allhelp = "\n".join(
        docstring
        + _inputs_help(cls)
        + [""]
        + _outputs_help(cls)
        + [""]
        + _refs_help(cls)
    )
    return allhelp.expandtabs(8)


def _inputs_help(cls):
    r"""
    Prints description for input parameters

    >>> from nipype.interfaces.afni import GCOR
    >>> _inputs_help(GCOR)  # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
    ['Inputs::', '', '\t[Mandatory]', '\tin_file: (a pathlike object or string...

    """
    helpstr = ["Inputs::"]
    mandatory_keys = []
    optional_items = []

    if cls.input_spec:
        inputs = cls.input_spec()
        mandatory_items = list(inputs.traits(mandatory=True).items())
        if mandatory_items:
            helpstr += ["", "\t[Mandatory]"]
            for name, spec in mandatory_items:
                helpstr += get_trait_desc(inputs, name, spec)

        mandatory_keys = {item[0] for item in mandatory_items}
        optional_items = [
            "\n".join(get_trait_desc(inputs, name, val))
            for name, val in inputs.traits(transient=None).items()
            if name not in mandatory_keys
        ]
        if optional_items:
            helpstr += ["", "\t[Optional]"] + optional_items

    if not mandatory_keys and not optional_items:
        helpstr += ["", "\tNone"]
    return helpstr


def _outputs_help(cls):
    r"""
    Prints description for output parameters

    >>> from nipype.interfaces.afni import GCOR
    >>> _outputs_help(GCOR)  # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
    ['Outputs::', '', '\tout: (a float)\n\t\tglobal correlation value']

    """
    helpstr = ["Outputs::", "", "\tNone"]
    if cls.output_spec:
        outputs = cls.output_spec()
        outhelpstr = [
            "\n".join(get_trait_desc(outputs, name, spec))
            for name, spec in outputs.traits(transient=None).items()
        ]
        if outhelpstr:
            helpstr = helpstr[:-1] + outhelpstr
    return helpstr


def _refs_help(cls):
    """Prints interface references."""
    references = getattr(cls, "references_", None)
    if not references:
        return []

    helpstr = ["References:", "-----------"]
    for r in references:
        helpstr += ["{}".format(r["entry"])]

    return helpstr


def get_trait_desc(inputs, name, spec):
    """Parses a HasTraits object into a nipype documentation string"""
    desc = spec.desc
    xor = spec.xor
    requires = spec.requires
    argstr = spec.argstr

    manhelpstr = ["\t%s" % name]

    type_info = spec.full_info(inputs, name, None)

    default = ""
    if spec.usedefault:
        default = ", nipype default value: %s" % str(spec.default_value()[1])
    line = "(%s%s)" % (type_info, default)

    manhelpstr = wrap(
        line,
        HELP_LINEWIDTH,
        initial_indent=manhelpstr[0] + ": ",
        subsequent_indent="\t\t  ",
    )

    if desc:
        for line in desc.split("\n"):
            line = re.sub(r"\s+", " ", line)
            manhelpstr += wrap(
                line, HELP_LINEWIDTH, initial_indent="\t\t", subsequent_indent="\t\t"
            )

    if argstr:
        pos = spec.position
        if pos is not None:
            manhelpstr += wrap(
                "argument: ``%s``, position: %s" % (argstr, pos),
                HELP_LINEWIDTH,
                initial_indent="\t\t",
                subsequent_indent="\t\t",
            )
        else:
            manhelpstr += wrap(
                "argument: ``%s``" % argstr,
                HELP_LINEWIDTH,
                initial_indent="\t\t",
                subsequent_indent="\t\t",
            )

    if xor:
        line = "%s" % ", ".join(xor)
        manhelpstr += wrap(
            line,
            HELP_LINEWIDTH,
            initial_indent="\t\tmutually_exclusive: ",
            subsequent_indent="\t\t  ",
        )

    if requires:
        others = [field for field in requires if field != name]
        line = "%s" % ", ".join(others)
        manhelpstr += wrap(
            line,
            HELP_LINEWIDTH,
            initial_indent="\t\trequires: ",
            subsequent_indent="\t\t  ",
        )
    return manhelpstr



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..calib import SFPICOCalibData


def test_SFPICOCalibData_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        info_file=dict(
            argstr="-infooutputfile %s",
            extensions=None,
            genfile=True,
            hash_files=False,
            mandatory=True,
        ),
        onedtfarange=dict(
            argstr="-onedtfarange %s",
            units="NA",
        ),
        onedtfastep=dict(
            argstr="-onedtfastep %f",
            units="NA",
        ),
        out_file=dict(
            argstr="> %s",
            extensions=None,
            genfile=True,
            position=-1,
        ),
        scheme_file=dict(
            argstr="-schemefile %s",
            extensions=None,
            mandatory=True,
        ),
        seed=dict(
            argstr="-seed %f",
            units="NA",
        ),
        snr=dict(
            argstr="-snr %f",
            units="NA",
        ),
        trace=dict(
            argstr="-trace %f",
            units="NA",
        ),
        twodtanglerange=dict(
            argstr="-twodtanglerange %s",
            units="NA",
        ),
        twodtanglestep=dict(
            argstr="-twodtanglestep %f",
            units="NA",
        ),
        twodtfarange=dict(
            argstr="-twodtfarange %s",
            units="NA",
        ),
        twodtfastep=dict(
            argstr="-twodtfastep %f",
            units="NA",
        ),
        twodtmixmax=dict(
            argstr="-twodtmixmax %f",
            units="NA",
        ),
        twodtmixstep=dict(
            argstr="-twodtmixstep %f",
            units="NA",
        ),
    )
    inputs = SFPICOCalibData.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_SFPICOCalibData_outputs():
    output_map = dict(
        PICOCalib=dict(
            extensions=None,
        ),
        calib_info=dict(
            extensions=None,
        ),
    )
    outputs = SFPICOCalibData.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..brains import BRAINSConstellationModeler


def test_BRAINSConstellationModeler_inputs():
    input_map = dict(
        BackgroundFillValue=dict(
            argstr="--BackgroundFillValue %s",
        ),
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        inputTrainingList=dict(
            argstr="--inputTrainingList %s",
            extensions=None,
        ),
        mspQualityLevel=dict(
            argstr="--mspQualityLevel %d",
        ),
        numberOfThreads=dict(
            argstr="--numberOfThreads %d",
        ),
        optimizedLandmarksFilenameExtender=dict(
            argstr="--optimizedLandmarksFilenameExtender %s",
        ),
        outputModel=dict(
            argstr="--outputModel %s",
            hash_files=False,
        ),
        rescaleIntensities=dict(
            argstr="--rescaleIntensities ",
        ),
        rescaleIntensitiesOutputRange=dict(
            argstr="--rescaleIntensitiesOutputRange %s",
            sep=",",
        ),
        resultsDir=dict(
            argstr="--resultsDir %s",
            hash_files=False,
        ),
        saveOptimizedLandmarks=dict(
            argstr="--saveOptimizedLandmarks ",
        ),
        trimRescaledIntensities=dict(
            argstr="--trimRescaledIntensities %f",
        ),
        verbose=dict(
            argstr="--verbose ",
        ),
        writedebuggingImagesLevel=dict(
            argstr="--writedebuggingImagesLevel %d",
        ),
    )
    inputs = BRAINSConstellationModeler.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_BRAINSConstellationModeler_outputs():
    output_map = dict(
        outputModel=dict(
            extensions=None,
        ),
        resultsDir=dict(),
    )
    outputs = BRAINSConstellationModeler.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Interfaces to functionality from nitime for time-series analysis of fmri data

- nitime.analysis.CoherenceAnalyzer: Coherence/y
- nitime.fmri.io:
- nitime.viz.drawmatrix_channels

"""
import numpy as np
import tempfile

from ...utils.misc import package_check
from ...utils.filemanip import fname_presuffix
from .base import NitimeBaseInterface
from ..base import (
    TraitedSpec,
    File,
    Undefined,
    traits,
    isdefined,
    BaseInterfaceInputSpec,
)

have_nitime = True
try:
    package_check("nitime")
except ImportError:
    have_nitime = False


class CoherenceAnalyzerInputSpec(BaseInterfaceInputSpec):

    # Input either csv file, or time-series object and use _xor_inputs to
    # discriminate
    _xor_inputs = ("in_file", "in_TS")
    in_file = File(
        desc=(
            "csv file with ROIs on the columns and "
            "time-points on the rows. ROI names at the top row"
        ),
        exists=True,
        requires=("TR",),
    )

    # If you gave just a file name, you need to specify the sampling_rate:
    TR = traits.Float(
        desc=("The TR used to collect the data in your csv file <in_file>")
    )

    in_TS = traits.Any(desc="a nitime TimeSeries object")

    NFFT = traits.Range(
        low=32,
        value=64,
        usedefault=True,
        desc=(
            "This is the size of the window used for "
            "the spectral estimation. Use values between "
            "32 and the number of samples in your time-series."
            "(Defaults to 64.)"
        ),
    )
    n_overlap = traits.Range(
        low=0,
        value=0,
        usedefault=True,
        desc=(
            "The number of samples which overlap"
            "between subsequent windows.(Defaults to 0)"
        ),
    )

    frequency_range = traits.List(
        value=[0.02, 0.15],
        usedefault=True,
        minlen=2,
        maxlen=2,
        desc=(
            "The range of frequencies over"
            "which the analysis will average."
            "[low,high] (Default [0.02,0.15]"
        ),
    )

    output_csv_file = File(
        desc="File to write outputs (coherence,time-delay) with file-names: "
        "``file_name_{coherence,timedelay}``"
    )

    output_figure_file = File(
        desc="""\
File to write output figures (coherence,time-delay) with file-names:
``file_name_{coherence,timedelay}``. Possible formats: .png,.svg,.pdf,.jpg,..."""
    )

    figure_type = traits.Enum(
        "matrix",
        "network",
        usedefault=True,
        desc=(
            "The type of plot to generate, where "
            "'matrix' denotes a matrix image and"
            "'network' denotes a graph representation."
            " Default: 'matrix'"
        ),
    )


class CoherenceAnalyzerOutputSpec(TraitedSpec):
    coherence_array = traits.Array(
        desc=("The pairwise coherence values between the ROIs")
    )

    timedelay_array = traits.Array(
        desc=("The pairwise time delays between the ROIs (in seconds)")
    )

    coherence_csv = File(desc=("A csv file containing the pairwise  coherence values"))

    timedelay_csv = File(desc=("A csv file containing the pairwise time delay values"))

    coherence_fig = File(desc=("Figure representing coherence values"))
    timedelay_fig = File(desc=("Figure representing coherence values"))


class CoherenceAnalyzer(NitimeBaseInterface):
    """Wraps nitime.analysis.CoherenceAnalyzer: Coherence/y"""

    input_spec = CoherenceAnalyzerInputSpec
    output_spec = CoherenceAnalyzerOutputSpec

    def _read_csv(self):
        """
        Read from csv in_file and return an array and ROI names

        The input file should have a first row containing the names of the
        ROIs (strings)

        the rest of the data will be read in and transposed so that the rows
        (TRs) will becomes the second (and last) dimension of the array

        """
        # Check that input conforms to expectations:
        first_row = open(self.inputs.in_file).readline()
        if not first_row[1].isalpha():
            raise ValueError(
                "First row of in_file should contain ROI names as strings of characters"
            )

        roi_names = (
            open(self.inputs.in_file).readline().replace('"', "").strip("\n").split(",")
        )
        # Transpose, so that the time is the last dimension:
        data = np.loadtxt(self.inputs.in_file, skiprows=1, delimiter=",").T

        return data, roi_names

    def _csv2ts(self):
        """ Read data from the in_file and generate a nitime TimeSeries object"""
        from nitime.timeseries import TimeSeries

        data, roi_names = self._read_csv()

        TS = TimeSeries(data=data, sampling_interval=self.inputs.TR, time_unit="s")

        TS.metadata = dict(ROIs=roi_names)

        return TS

    # Rewrite _run_interface, but not run
    def _run_interface(self, runtime):
        import nitime.analysis as nta

        lb, ub = self.inputs.frequency_range

        if self.inputs.in_TS is Undefined:
            # get TS form csv and inputs.TR
            TS = self._csv2ts()

        else:
            # get TS from inputs.in_TS
            TS = self.inputs.in_TS

        # deal with creating or storing ROI names:
        if "ROIs" not in TS.metadata:
            self.ROIs = ["roi_%d" % x for x, _ in enumerate(TS.data)]
        else:
            self.ROIs = TS.metadata["ROIs"]

        A = nta.CoherenceAnalyzer(
            TS,
            method=dict(
                this_method="welch",
                NFFT=self.inputs.NFFT,
                n_overlap=self.inputs.n_overlap,
            ),
        )

        freq_idx = np.where(
            (A.frequencies > self.inputs.frequency_range[0])
            * (A.frequencies < self.inputs.frequency_range[1])
        )[0]

        # Get the coherence matrix from the analyzer, averaging on the last
        # (frequency) dimension: (roi X roi array)
        self.coherence = np.mean(A.coherence[:, :, freq_idx], -1)
        # Get the time delay from analyzer, (roi X roi array)
        self.delay = np.mean(A.delay[:, :, freq_idx], -1)
        return runtime

    # Rewrite _list_outputs (look at BET)
    def _list_outputs(self):
        outputs = self.output_spec().get()

        # if isdefined(self.inputs.output_csv_file):

        # write to a csv file and assign a value to self.coherence_file (a
        # file name + path)

        # Always defined (the arrays):
        outputs["coherence_array"] = self.coherence
        outputs["timedelay_array"] = self.delay

        # Conditional
        if isdefined(self.inputs.output_csv_file) and hasattr(self, "coherence"):
            # we need to make a function that we call here that writes the
            # coherence values to this file "coherence_csv" and makes the
            # time_delay csv file??
            self._make_output_files()
            outputs["coherence_csv"] = fname_presuffix(
                self.inputs.output_csv_file, suffix="_coherence"
            )

            outputs["timedelay_csv"] = fname_presuffix(
                self.inputs.output_csv_file, suffix="_delay"
            )

        if isdefined(self.inputs.output_figure_file) and hasattr(self, "coherence"):
            self._make_output_figures()
            outputs["coherence_fig"] = fname_presuffix(
                self.inputs.output_figure_file, suffix="_coherence"
            )
            outputs["timedelay_fig"] = fname_presuffix(
                self.inputs.output_figure_file, suffix="_delay"
            )

        return outputs

    def _make_output_files(self):
        """
        Generate the output csv files.
        """
        for this in zip([self.coherence, self.delay], ["coherence", "delay"]):
            tmp_f = tempfile.mkstemp()[1]
            np.savetxt(tmp_f, this[0], delimiter=",")

            fid = open(
                fname_presuffix(self.inputs.output_csv_file, suffix="_%s" % this[1]),
                "w+",
            )
            # this writes ROIs as header line
            fid.write("," + ",".join(self.ROIs) + "\n")
            # this writes ROI and data to a line
            for r, line in zip(self.ROIs, open(tmp_f)):
                fid.write("%s,%s" % (r, line))
            fid.close()

    def _make_output_figures(self):
        """
        Generate the desired figure and save the files according to
        self.inputs.output_figure_file

        """
        import nitime.viz as viz

        if self.inputs.figure_type == "matrix":
            fig_coh = viz.drawmatrix_channels(
                self.coherence, channel_names=self.ROIs, color_anchor=0
            )

            fig_coh.savefig(
                fname_presuffix(self.inputs.output_figure_file, suffix="_coherence")
            )

            fig_dt = viz.drawmatrix_channels(
                self.delay, channel_names=self.ROIs, color_anchor=0
            )

            fig_dt.savefig(
                fname_presuffix(self.inputs.output_figure_file, suffix="_delay")
            )
        else:
            fig_coh = viz.drawgraph_channels(self.coherence, channel_names=self.ROIs)

            fig_coh.savefig(
                fname_presuffix(self.inputs.output_figure_file, suffix="_coherence")
            )

            fig_dt = viz.drawgraph_channels(self.delay, channel_names=self.ROIs)

            fig_dt.savefig(
                fname_presuffix(self.inputs.output_figure_file, suffix="_delay")
            )



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
The workbench module provides classes for interfacing with `connectome workbench
<https://www.humanconnectome.org/software/connectome-workbench>`_ tools.

Connectome Workbench is an open source, freely available visualization and
discovery tool used to map neuroimaging data, especially data generated by the
Human Connectome Project.
"""

import os
import re

from ... import logging
from ...utils.filemanip import split_filename
from ..base import CommandLine, PackageInfo

iflogger = logging.getLogger("nipype.interface")


class Info(PackageInfo):
    """Handle Connectome Workbench version information."""

    version_cmd = "wb_command -version"

    @staticmethod
    def parse_version(raw_info):
        m = re.search(r"\nVersion (\S+)", raw_info)
        return m.groups()[0] if m else None


class WBCommand(CommandLine):
    """Base support for workbench commands."""

    @property
    def version(self):
        return Info.version()

    def _gen_filename(self, name, outdir=None, suffix="", ext=None):
        """Generate a filename based on the given parameters.
        The filename will take the form: <basename><suffix><ext>.
        Parameters
        ----------
        name : str
            Filename to base the new filename on.
        suffix : str
            Suffix to add to the `basename`.  (defaults is '' )
        ext : str
            Extension to use for the new filename.
        Returns
        -------
        fname : str
            New filename based on given parameters.
        """
        if not name:
            raise ValueError("Cannot generate filename - filename not set")

        _, fname, fext = split_filename(name)
        if ext is None:
            ext = fext
        if outdir is None:
            outdir = "."
        return os.path.join(outdir, fname + suffix + ext)



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..preprocess import DARTELNorm2MNI


def test_DARTELNorm2MNI_inputs():
    input_map = dict(
        apply_to_files=dict(
            copyfile=False,
            field="mni_norm.data.subjs.images",
            mandatory=True,
        ),
        bounding_box=dict(
            field="mni_norm.bb",
        ),
        flowfield_files=dict(
            field="mni_norm.data.subjs.flowfields",
            mandatory=True,
        ),
        fwhm=dict(
            field="mni_norm.fwhm",
        ),
        matlab_cmd=dict(),
        mfile=dict(
            usedefault=True,
        ),
        modulate=dict(
            field="mni_norm.preserve",
        ),
        paths=dict(),
        template_file=dict(
            copyfile=False,
            extensions=[".hdr", ".img", ".img.gz", ".nii"],
            field="mni_norm.template",
            mandatory=True,
        ),
        use_mcr=dict(),
        use_v8struct=dict(
            min_ver="8",
            usedefault=True,
        ),
        voxel_size=dict(
            field="mni_norm.vox",
        ),
    )
    inputs = DARTELNorm2MNI.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_DARTELNorm2MNI_outputs():
    output_map = dict(
        normalization_parameter_file=dict(
            extensions=None,
        ),
        normalized_files=dict(),
    )
    outputs = DARTELNorm2MNI.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..base import Split


def test_Split_inputs():
    input_map = dict(
        inlist=dict(
            mandatory=True,
        ),
        splits=dict(
            mandatory=True,
        ),
        squeeze=dict(
            usedefault=True,
        ),
    )
    inputs = Split.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_Split_outputs():
    output_map = dict()
    outputs = Split.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
"""Parallel workflow execution via PBS/Torque
"""

import os
import sys

from .base import GraphPluginBase, logger

soma_not_loaded = False
try:
    from soma.workflow.client import Job, Workflow, WorkflowController, Helper
except:
    soma_not_loaded = True


class SomaFlowPlugin(GraphPluginBase):
    """Execute using Soma workflow
    """

    def __init__(self, plugin_args=None):
        if soma_not_loaded:
            raise ImportError("SomaFlow could not be imported")
        super(SomaFlowPlugin, self).__init__(plugin_args=plugin_args)

    def _submit_graph(self, pyfiles, dependencies, nodes):
        jobs = []
        soma_deps = []
        for idx, fname in enumerate(pyfiles):
            name = os.path.splitext(os.path.split(fname)[1])[0]
            jobs.append(Job(command=[sys.executable, fname], name=name))
        for key, values in list(dependencies.items()):
            for val in values:
                soma_deps.append((jobs[val], jobs[key]))

        wf = Workflow(jobs, soma_deps)
        logger.info("serializing workflow")
        Helper.serialize("workflow", wf)
        controller = WorkflowController()
        logger.info("submitting workflow")
        wf_id = controller.submit_workflow(wf)
        Helper.wait_workflow(wf_id, controller)



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..utils import MRIsConvert


def test_MRIsConvert_inputs():
    input_map = dict(
        annot_file=dict(
            argstr="--annot %s",
            extensions=None,
        ),
        args=dict(
            argstr="%s",
        ),
        dataarray_num=dict(
            argstr="--da_num %d",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        functional_file=dict(
            argstr="-f %s",
            extensions=None,
        ),
        in_file=dict(
            argstr="%s",
            extensions=None,
            mandatory=True,
            position=-2,
        ),
        label_file=dict(
            argstr="--label %s",
            extensions=None,
        ),
        labelstats_outfile=dict(
            argstr="--labelstats %s",
            extensions=None,
        ),
        normal=dict(
            argstr="-n",
        ),
        origname=dict(
            argstr="-o %s",
        ),
        out_datatype=dict(
            mandatory=True,
            xor=["out_file"],
        ),
        out_file=dict(
            argstr="%s",
            extensions=None,
            genfile=True,
            mandatory=True,
            position=-1,
            xor=["out_datatype"],
        ),
        parcstats_file=dict(
            argstr="--parcstats %s",
            extensions=None,
        ),
        patch=dict(
            argstr="-p",
        ),
        rescale=dict(
            argstr="-r",
        ),
        scalarcurv_file=dict(
            argstr="-c %s",
            extensions=None,
        ),
        scale=dict(
            argstr="-s %.3f",
        ),
        subjects_dir=dict(),
        talairachxfm_subjid=dict(
            argstr="-t %s",
        ),
        to_scanner=dict(
            argstr="--to-scanner",
        ),
        to_tkr=dict(
            argstr="--to-tkr",
        ),
        vertex=dict(
            argstr="-v",
        ),
        xyz_ascii=dict(
            argstr="-a",
        ),
    )
    inputs = MRIsConvert.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_MRIsConvert_outputs():
    output_map = dict(
        converted=dict(
            extensions=None,
        ),
    )
    outputs = MRIsConvert.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from nipype.testing import example_data


def test_split_and_merge(tmpdir):
    import numpy as np
    import nibabel as nb
    import os.path as op
    import os

    from nipype.algorithms.misc import split_rois, merge_rois

    in_mask = example_data("tpms_msk.nii.gz")
    dwfile = tmpdir.join("dwi.nii.gz").strpath
    mask_img = nb.load(in_mask)
    mskdata = np.asanyarray(mask_img.dataobj)
    aff = mask_img.affine

    dwshape = (mskdata.shape[0], mskdata.shape[1], mskdata.shape[2], 6)
    dwdata = np.random.normal(size=dwshape)
    tmpdir.chdir()
    nb.Nifti1Image(dwdata.astype(np.float32), aff, None).to_filename(dwfile)

    resdw, resmsk, resid = split_rois(dwfile, in_mask, roishape=(20, 20, 2))
    merged = merge_rois(resdw, resid, in_mask)
    dwmerged = nb.load(merged).get_fdata(dtype=np.float32)

    dwmasked = dwdata * mskdata[:, :, :, np.newaxis]

    assert np.allclose(dwmasked, dwmerged)
